import functools

from .meta import SerializableMeta, AttributedMeta
from ..registry import ModelRegistry
from .mixin import Importable, NewMixin, Exportable
from .utils import _collect_attrs, _schema_cls_from_attributed_class
from onto.mapper.schema import Schema


class SimpleStore:
    """
    To store simple business properties
    """
    pass


class PonyStore:

    def _set_owner(self, owner):
        self.parent = owner

    def __getattr__(self, item):
        _self = self.parent
        attr = _self._adict_[item]
        inner = _self._vals_
        return inner[attr]


class AttributedMixin(metaclass=AttributedMeta):
    """ helper to declare mixin classes with attributes
    Example:
        class LatLonMixin(AttributedMixin):
            lat = attrs.float
            lon = attrs.float

        class Location(LatLonMixin, DomainModel):
            name = attrs.string

        # Location will have attributes lat, lon, and name.

    """
    @classmethod
    def attrs(cls):
        return cls.__attributes


class BaseRegisteredModel(metaclass=ModelRegistry):
    """
    Ref: https://github.com/faif/python-patterns/blob/master/patterns/behavioral/registry__py3.py
    """

    @classmethod
    def _get_children(cls):
        return {cls.get_cls_from_name(c_str)
                for c_str in cls._get_children_str(cls.__name__)}

    @classmethod
    def _get_subclasses(cls):
        res = {cls, }
        children = cls._get_children()
        for child in children:
            res |= child._get_subclasses()
        return res

    @classmethod
    def _get_subclasses_str(cls):
        return list(
            sorted(_cls.__name__ for _cls in cls._get_subclasses())
        )

    @classmethod
    def _get_parents(cls):
        return {cls.get_cls_from_name(c_str)
                for c_str in cls._get_parents_str(cls.__name__)}

    get_parents = _get_parents
    get_subclasses = _get_subclasses


class SchemedBase:

    @property
    def schema_obj(self):
        raise NotImplementedError

    @property
    def schema_cls(self):
        raise NotImplementedError


class Schemed(SchemedBase):
    """
    A mixin class for object bounded to a schema for serialization
        and deserialization.
    TODO: maybe add caching
    Currently schema is reconstructed at every call to
        refresh value for different contexts.
    """

    _schema_obj = None
    _schema_cls = None

    # def __init__(self, *args, **kwargs):
    #     super().__init__(*args, **kwargs)
    # self._schema_obj = self._schema_cls()

    @classmethod
    @functools.lru_cache(maxsize=None)
    def get_schema_cls(cls):
        """ Returns the Schema class associated with the model class.
        """
        return _schema_cls_from_attributed_class(cls=cls)

    @classmethod
    def get_schema_obj(cls):
        """ Returns an instantiated object for Schema associated
                with the model class
        """
        # Use __dict__ to avoid reading from super class
        # if "_schema_obj" not in cls.__dict__:
        #     schema_cls = cls.get_schema_cls()
        #     if schema_cls is None:
        #         return None
        #     cls._schema_obj = schema_cls()
        # return cls._schema_obj
        if _schema_cls := cls.get_schema_cls():
            _schema_obj = _schema_cls()
            return _schema_obj

    @property
    def schema_cls(self):
        """ Returns the Schema class associated with the model object.
        """
        return self.get_schema_cls()

    @property
    def schema_obj(self):
        """ Returns an instantiated object for Schema associated
                with the model object.
        """
        return self.get_schema_obj()

    @classmethod
    def _get_fields(cls):
        fd = cls.get_schema_obj().fields
        return {
            key: val for key, val in fd.items()  # TODO: change
        }

    #     """ TODO: find ways of collecting fields without reading
    #                 private attribute on Marshmallow.Schema
    #
    #     :return:
    #     """
    #     res = dict()
    #     for name, declared_field in cls.get_schema_obj().fields.items():
    #         if not declared_field.dump_only:
    #             res[name] = declared_field
    #     return res


class Mutable(BaseRegisteredModel,
              Schemed, Importable, NewMixin, Exportable):
    pass


class Immutable(BaseRegisteredModel, Schemed, NewMixin, Exportable):
    pass


class GraphqlMixin(BaseRegisteredModel):

    _union = False

    def graphql_field_resolve(self, info: 'GraphQLResolveInfo', *args):
        from graphql import GraphQLResolveInfo
        info: GraphQLResolveInfo
        field_name = info.field_name
        return self.graphql_representation.get(field_name, None)

    import functools

    @property
    @functools.lru_cache(maxsize=None)
    def graphql_representation(self):
        return self.to_dict()

    @staticmethod
    def _convert(key, attr, is_input):
        if not attr.properties.is_internal:
            from onto.attrs.unit import MonadContext
            context = MonadContext.context()
            context = context.graphql_capable(is_input=is_input)
            context = context.attribute_name(key)

            annotation = attr.annotation
            context = context.annotate(annotation)

            context = context.data_key_from_name()
            context = context.optional

            with context:
                return attr.properties.data_key, attr.graphql_field

    @classmethod
    def __fields_gen(cls, is_input):
        for key, attr in _collect_attrs(cls):
            # TODO: check to see if .attribute_name(key).data_key(key) should be removed
            if ret := cls._convert(key=key, attr=attr, is_input=is_input):
                yield ret
        else:
            yield from ()

    @classmethod
    def _get_union_graphql_interface(cls):
        """
        本层
        """
        import graphql

        def resolve_type(__self, value, _type):
            raise NotImplementedError

        fields = lambda: dict(cls.__fields_gen(is_input=is_input))
        type_name = cls.__name__

        ot = graphql.GraphQLInterfaceType(
            name=type_name,
            resolve_type=resolve_type,
            fields=fields,
        )
        return ot

    @classmethod
    def get_graphql_interfaces(cls):
        for parent in cls.get_parents():
            if issubclass(parent, GraphqlMixin):
                yield from parent.get_graphql_interfaces()

        if cls._union:
            yield cls._get_union_graphql_interface()

        yield from ()  # Empty case

    @classmethod
    @functools.lru_cache(maxsize=None)
    def get_graphql_object_type(cls, is_input=False):

        if unwrap := getattr(cls.Meta, "unwrap", None):

            key, attr = next((k, v) for k, v in _collect_attrs(cls) if k == unwrap)
            data_key, graphql_field = cls._convert(key=key, attr=attr, is_input=is_input)

            from graphql import GraphQLScalarType
            assert isinstance(graphql_field.type, GraphQLScalarType)

            this_type = GraphQLScalarType(
                cls.__name__,
                serialize=graphql_field.type.serialize,
                parse_value=graphql_field.type.parse_value,
                parse_literal=graphql_field.type.parse_literal,
            )
            return this_type
        else:
            import graphql
            if not is_input:
                base = graphql.GraphQLObjectType
                type_name = cls.__name__
            else:
                base = graphql.GraphQLInputObjectType
                type_name = f'{cls.__name__}Input'

            fields = lambda: dict(cls.__fields_gen(is_input=is_input))

            ot = base(
                type_name,
                fields=fields
            )
            return ot


class Serializable(Mutable, GraphqlMixin, metaclass=SerializableMeta):

    class Meta:
        pass
    #     """
    #     Options object for a Serializable model.
    #     """
    #     schema_cls = None

    _schema_base = Schema

    def _init__attrs(self):
        if not getattr(self, '_attrs', None):
            self._attrs = SimpleStore()
        from onto.attrs.unit import MonadContext
        with MonadContext.context().init_options(initialize=False, initializer=None):
            for key, attr in _collect_attrs(cls=self.__class__):
                if attr.properties.initialize:
                    initializer = attr.properties.make_init(name=key)
                    initializer(self)

    def __init__(self, *args, **kwargs):
        self._init__attrs()
        super().__init__(*args, **kwargs)



import pytest

from examples.meeting_room import domain_models


@pytest.fixture
def users(request, CTX):
    tijuana = domain_models.User.new(doc_id="tijuana")
    tijuana.first_name = "Tijuana"
    tijuana.last_name = "Furlong"
    tijuana.hearing_aid_requested = True
    tijuana.organization = "UCSD"
    print(tijuana.to_dict())
    tijuana.save()

    thomasina = domain_models.User.new(doc_id="thomasina")
    thomasina.first_name = "Thomasina"
    thomasina.last_name = "Manes"
    thomasina.hearing_aid_requested = False
    thomasina.organization = "UCSD"
    thomasina.save()

    joshua = domain_models.User.new(doc_id="joshua")
    joshua.first_name = "Joshua"
    joshua.last_name = "Pendergrast"
    joshua.hearing_aid_requested = True
    joshua.organization = "SDSU"
    joshua.save()

    def fin():
        tijuana.delete()
        thomasina.delete()
        joshua.delete()

    request.addfinalizer(fin)

    return [tijuana, thomasina, joshua]


@pytest.fixture
def tickets(users, request, CTX):
    tj_t = domain_models.Ticket.new(doc_id="tj_t")
    tj_t.role = "Participant"
    tj_t.user = domain_models.User._get_collection().child("tijuana")
    tj_t.attendance = True
    tj_t.save()

    ts_t = domain_models.Ticket.new(doc_id="ts_t")
    ts_t.role = "Organizer"
    ts_t.user = domain_models.User._get_collection().child("thomasina")
    ts_t.attendance = True
    ts_t.save()

    js_t = domain_models.Ticket.new(doc_id="js_t")
    js_t.role = "Participant"
    js_t.user = domain_models.User._get_collection().child("joshua")
    js_t.attendance = True
    js_t.save()

    def fin():
        tj_t.delete()
        ts_t.delete()
        js_t.delete()

    request.addfinalizer(fin)

    return [tj_t, ts_t, js_t]


@pytest.fixture
def location(request, CTX):
    obj = domain_models.Location.new()
    obj.latitude = 32.880361
    obj.longitude = -117.242929
    obj.address = "9500 Gilman Drive, La Jolla, CA"

    obj.save()

    def fin():
        obj.delete()

    request.addfinalizer(fin)

    return obj


@pytest.fixture
def meeting(users, tickets, location, request, CTX):
    m = domain_models.Meeting.new(doc_id="meeting_1")
    m.users = [user.doc_ref for user in users]
    m.tickets = {ticket.user.id: ticket.doc_ref for ticket in tickets}
    m.location = location.doc_ref
    m.status = "in-session"
    m.save()

    def fin():
        m.delete()

    request.addfinalizer(fin)

    return m



import pytest


def test_connector():
    from onto.connector.base import ConnectorBase

    commands = list()

    class M:

        source = ConnectorBase()

        @source.before_call
        def before(self):
            commands.append('before1')

        @source.after_call
        def after(self):
            commands.append('after1')

    assert not M().source.invoke_mediator('hello')
    assert commands == ['before1', 'after1']



from google.cloud.firestore_v1 import DocumentSnapshot

from examples.city.forms import CityForm
from examples.city.views import CityView
from onto.view.base import ViewMediatorBase
from onto.view_mediator_dav import ViewMediatorDeltaDAV, ProtocolBase


class CityViewMediator(ViewMediatorDeltaDAV):

    def notify(self, obj):
        obj.save()

    class Protocol(ProtocolBase):

        @staticmethod
        def on_create(snapshot: DocumentSnapshot, mediator: ViewMediatorBase):
            view = CityView.new(snapshot=snapshot)
            mediator.notify(obj=view)


class CityFormMediator(ViewMediatorDeltaDAV):

    def notify(self, obj):
        obj.propagate_change()

    class Protocol(ProtocolBase):

        @staticmethod
        def on_create(snapshot: DocumentSnapshot, mediator: ViewMediatorBase):
            obj = CityForm.new(doc_ref=snapshot.reference)
            obj.update_vals(with_raw=snapshot.to_dict())
            mediator.notify(obj=obj)



from onto.database.firestore import TargetIdAssigner
from onto.query.query import DomainModelQuery


def test_reference():

    from onto.database import Reference

    r = Reference()
    a = r.child('a')
    ab = a.child('b')
    assert a == 'a'
    assert ab == 'a/b'


def test_reference_truediv():

    from onto.database import Reference

    r = Reference()
    a = r / 'a'
    ab = a / 'b'
    assert a == 'a'
    assert ab == 'a/b'

    assert a/'c'/'d' == 'a/c/d'
    assert a/'e'/'f'/'g' == 'a/e/f/g'


def test_reference_truediv_eq():
    """ Tests that "/=" creates a new Reference object
    """

    from onto.database import Reference

    r = Reference()
    r /= 'a'
    pre = id(r)
    assert r == 'a'
    r /= 'b'
    post = id(r)
    assert r == 'a/b'

    assert pre != post


def test_deserialize():

    from onto.database import Reference

    r = Reference.from_str(s='a/b')

    assert r == 'a/b'
    assert isinstance(r, Reference)
    assert not r._is_empty

    r = Reference.from_str(s='')
    assert r == ''
    assert isinstance(r, Reference)
    assert r._is_empty

    # This behavior is arbitrary and may be changed
    r = Reference.from_str(s='/')
    assert r == '/'
    assert isinstance(r, Reference)
    assert not r._is_empty


def test_serialize():

    from onto.database import Reference

    r = Reference()
    a = r.child('a')
    ab = a.child('b')
    assert str(a) == 'a'
    assert str(ab) == 'a/b'


def test_leancloud():
    from onto.database.leancloud import LeancloudDatabase
    from onto.database import Snapshot

    ref = LeancloudDatabase.ref/'TODO'/'582570f38ac247004f39c24b'
    snapshot = Snapshot(title='foo', priority='bar')
    LeancloudDatabase.set(ref=ref, snapshot=snapshot)


def test_target_id_assigner():
    assigner = TargetIdAssigner()
    assert assigner.assign_id() == 32
    assert assigner.assign_id() == 33
    assigner.release_id(32)
    assert assigner.assign_id() == 32
    assert assigner.assign_id() == 34
    assert assigner.assign_id() == 35


def test_watch():
    from onto.watch import _Watch
    from onto.context import Context as CTX
    from google.cloud import firestore

    def callback(*args, **kwargs):
        print(f"{args} {kwargs}")

    _watch = _Watch(
        firestore=CTX.db.firestore_client,
        comparator=lambda d1, d2: 1,
        # snapshot_callback=callback,
        document_snapshot_cls=firestore.DocumentSnapshot,
        document_reference_cls=firestore.DocumentReference,
    )

    from google.cloud import firestore
    t1_ref = CTX.db.firestore_client.document('s/t1')
    t2_ref = CTX.db.firestore_client.document('s/t2')
    t1_ref.set({'foo': 'bar'})
    t2_ref.set({'foo': 'bar'})
    document_refs = [t1_ref._document_path, t2_ref._document_path]

    target = {
        "documents": {"documents": document_refs},
        "target_id": 0x2,
        "once": False
    }

    _watch.add_target(target, callback)

    from onto import testing_utils

    testing_utils._wait()
    # TODO: add and test tearDown for _watch
    # Implement close() for listener

def test_listener():
    from onto.database.firestore import FirestoreListener, Query
    from onto.domain_model import DomainModel
    from onto import attrs

    class S(DomainModel):

        foo = attrs.bproperty()
        obj_type = attrs.object_type(import_enabled=False, export_enabled=False)

        class Meta:
            collection_name = 'S'

    S.new(doc_id='T', foo='bar').save()

    query = DomainModelQuery(parent=S, arguments=[])

    from onto.source.firestore import FirestoreSource
    class M:
        source = FirestoreSource(query=query)

        @staticmethod
        @source.triggers.on_create
        def add_t(ref, snapshot):
            print(f"{ref} {snapshot}")

        @classmethod
        def start(cls):
            cls.source.start()

    M.start()

    from onto import testing_utils
    testing_utils._wait()

    testing_utils._wait()


def test_snapshot_init_meta():
    from onto.database.firestore import FirestoreSnapshot
    snapshot = FirestoreSnapshot.from_data_and_meta(
        data=dict(foo='bar'),
        my_arg=1
    )
    assert snapshot['foo'] == 'bar'
    assert 'my_arg' not in snapshot
    assert snapshot.my_arg == 1


# def test_stuffs():
#
#     from collections import UserDict
#
#     class ExpD(UserDict):
#         def __getitem__(self, item):
#             print(item)
#
#     ExpD()[1:2]



from typing import TypeVar, Type, Union

from onto.models.base import BaseRegisteredModel, Serializable, Schemed
from onto.models.mixin import Importable, Exportable, NewMixin

T = TypeVar('T')
U = TypeVar('U')


class ClsFactory:

    @classmethod
    def create_from_root(
            cls, name, schema,
            root_cls=None,
            additional_base_tuple=None):
        pass

    @classmethod
    def create_customized(
            cls, name, schema,
            importable=True,
            exportable=True):
        """

        :param name: name of the new class
        :param schema: schema of the new class
        :param auto_initialized: If set to True, the instance variables
                will be initialized to the default value read from
                the corresponding field.
        :param importable: If set to True, the instance variables can
                be set in batch from dictionary.
        :param exportable: If set to True, the object can be exported
                to a dictionary.
        :return:
        """
        base_list = [BaseRegisteredModel, Schemed]

        if importable:
            base_list.append(Importable)

        if exportable:
            base_list.append(Exportable)

        base_list.append(NewMixin)

        return cls.create(
            name=name,
            schema=schema,
            base_tuple=tuple(base_list)
        )

    @classmethod
    def create(cls, name, schema: Type[T],
               base: Type[U] = Serializable,
               base_tuple=None) -> Union[Type[U], Type[T]]:

        existing = BaseRegisteredModel.get_cls_from_name(name)
        if existing is None:
            new_cls = type(name,  # class name
                           base_tuple or (base, ),
                           dict(
                               _schema_cls=schema,
                           )
                           )
            return new_cls
        else:
            return existing



"""

Contains test cases inspired by flasgger (MIT licence)
See: https://github.com/flasgger/flasgger/blob/master/LICENSE
"""
import time

from flask import Flask, jsonify
from flasgger import Swagger, SwaggerView
from google.cloud.firestore import DocumentReference

from onto.view import GenericView
from onto import fields
from onto.schema import Schema
from onto.view_model import ViewModel
from onto.domain_model import DomainModel
from google.cloud import firestore
from onto import testing_utils
from functools import partial

from examples.luggage_models import LuggageItem, Luggages

from onto.config import Config
from onto.context import Context as CTX

if __name__ == "__main__":
    config = Config(
        app_name="flask-boiler-testing",
        debug=True,
        testing=True,
        certificate_filename="flask-boiler-testing-firebase-adminsdk-4m0ec-7505aaef8d.json"
    )

    CTX.read(config)

    d_a = {
            "luggage_type": "large",
            "weight_in_lbs": 20,
            "obj_type": "LuggageItem"
        }
    id_a = "luggage_id_a"
    obj_a = LuggageItem.new(id_a)
    obj_a._import_properties(d_a)
    obj_a.save()

    d_b = {
            "luggage_type": "medium",
            "weight_in_lbs": 15,
            "obj_type": "LuggageItem"
        }
    id_b = "luggage_id_b"
    obj_b = LuggageItem.new(id_b)
    obj_b._import_properties(d_b)
    obj_b.save()

    vm_ref: DocumentReference = CTX.db.document("test_lugagges/user_a_luggages")

    vm: Luggages = Luggages.new(vm_ref)

    vm.bind_to(key=id_a, obj_type="LuggageItem", doc_id=id_a)
    vm.bind_to(key=id_b, obj_type="LuggageItem", doc_id=id_b)
    vm.register_listener()

    # Takes time to propagate changes
    testing_utils._wait(factor=.5)

    assert vm.to_dict() == {
        "luggages": [
            {
                "luggage_type": "large",
                "weight_in_lbs": 20
            },
            {
                "luggage_type": "medium",
                "weight_in_lbs": 15
            }
        ],
        "total_weight": 35,
        "total_count": 2,
        # "obj_type": "Luggages"
    }

    # Change the weight on one of the luggages
    obj_b.weight_in_lbs = 25
    obj_b.save()

    # Note that the update is eventually consistent with delays
    #   and local copies of a ViewModel should not be used to read
    #   updated values

    testing_utils._wait(factor=.3)

    # # Test that the view model now has updated values
    assert vm.to_dict() == {
        "luggages": [
            {
                "luggage_type": "large",
                "weight_in_lbs": 20
            },
            {
                "luggage_type": "medium",
                "weight_in_lbs": 25
            }
        ],
        "total_weight": 45,
        "total_count": 2
    }



from onto.attrs.attribute import ReferenceAttribute
reference = ReferenceAttribute
from .store import Store
from .gallery import Gallery
from .struct import struct_ref
from .snapshot_container import SnapshotContainer
from .business_property_store import to_ref



import asyncio
import functools
from threading import Thread
from queue import Queue
from typing import Optional, Awaitable

from google.cloud.firestore_v1 import DocumentSnapshot, Watch, \
    DocumentReference

from onto import sink
from onto.watch import DocumentChange

from onto.context import Context as CTX
# # https://dev.to/googlecloud/portable-cloud-functions-with-the-python-functions-framework-a6a
# from google.cloud.functions.context import Context as GcfContext

from onto.query import run_transaction
from onto.view.base import ViewMediatorBase

# NOTE: note for deploying to cloud functions
# NOTE: gcloud functions deploy to_trigger --runtime python37 --trigger-event providers/cloud.firestore/eventTypes/document.create --trigger-resource "projects/flask-boiler-testing/databases/(default)/documents/gcfTest/{gcfTestDocId}"
# NOTE: No to unauthenticated invocations
from onto.watch import DataListener

EVENT_TYPE_MAPPING = dict(
    create="providers/cloud.firestore/eventTypes/document.create",
    update="providers/cloud.firestore/eventTypes/document.update",
    delete="providers/cloud.firestore/eventTypes/document.delete",
    write="providers/cloud.firestore/eventTypes/document.write"
)


class DeltaSink(sink.firestore):

    def emit(self, doc_ref, obj):
        doc_ref.save(obj.to_dict())


class ViewMediatorDeltaDAV(ViewMediatorBase):

    def notify(self, obj):
        """ Called when the object has state changes to be notified.

        :param obj: the object with state change
        :return:
        """
        self.sink.emit(obj.doc_ref, obj)

    # def __init_subclass__(cls, **kwargs):
    #     super().__init_subclass__(**kwargs)
    #     if not hasattr(cls, "Protocol"):
    #         raise NotImplementedError

    def __init__(self, *args, query, **kwargs):
        """ Initializes a ViewMediator to declare protocols that
                are called when the results of a query change. Note that
                mediator.start must be called later.

        :param args:
        :param query: a listener will be attached to this query
        :param kwargs:
        """
        super().__init__(*args, **kwargs)
        self.query = query
        self.sink = DeltaSink()

    def _on_snapshot(self, snapshots, changes, timestamp):
        """ For use with
        Note that server reboot will result in some "Modified" objects
            to be routed as "Added" objects

        :param snapshots:
        :param changes:
        :param timestamp:
        """
        for change in changes:
            snapshot = change.document
            assert isinstance(snapshot, DocumentSnapshot)
            try:
                CTX.logger.info(f"DAV: {self.__class__.__name__} started "
                                f"for {snapshot.reference.path}")
                if change.type.name == 'ADDED':
                    self.Protocol.on_create(
                        snapshot=snapshot,
                        mediator=self
                    )
                elif change.type.name == 'MODIFIED':
                    self.Protocol.on_update(
                        snapshot=snapshot,
                        mediator=self
                    )
                elif change.type.name == 'REMOVED':
                    self.Protocol.on_delete(
                        snapshot=snapshot,
                        mediator=self
                    )
            except Exception as e:
                """
                Expects e to be printed to the logger 
                """
                CTX.logger.exception(f"DAV {self.__class__.__name__} failed "
                                     f"for {snapshot.reference.path}")
            else:
                CTX.logger.info(f"DAV {self.__class__.__name__} succeeded "
                                f"or enqueued "
                                f"for {snapshot.reference.path}")

    def _get_on_snapshot(self):
        return self._on_snapshot

    def start(self):
        """ Starts a listener to the query.
        Do not use this for cloud functions.

        """

        query, on_snapshot = self.query, self._get_on_snapshot()

        self.listener = DataListener(snapshot_callback=on_snapshot,
                         firestore=CTX.db, once=False, query=query)


class ProtocolBase:
    """ Protocol to specify actions when a document is 'ADDED', 'MODIFIED', and
            'REMOVED'.

    """

    @staticmethod
    def on_create(snapshot: DocumentSnapshot, mediator: ViewMediatorBase):
        """
        Called when a document is 'ADDED' to the result of a query.

        :param snapshot: Firestore Snapshot added
        :param mediator: ViewMediator object
        """
        pass

    @staticmethod
    def on_update(snapshot: DocumentSnapshot, mediator: ViewMediatorBase):
        """ Called when a document is 'MODIFIED' in the result of a query.

        :param snapshot: Firestore Snapshot modified
        :param mediator: ViewMediator object
        """
        pass

    @staticmethod
    def on_delete(snapshot: DocumentSnapshot, mediator: ViewMediatorBase):
        """ Called when a document is 'REMOVED' in the result of a query.

        :param snapshot: Firestore Snapshot deleted
        :param mediator: ViewMediator object
        :return:
        """
        pass


class OnTriggerMixin:
    TRIGGER_EVENT_TYPE: str = "providers/cloud.firestore/eventTypes/document.write"

    def __init__(self, *args, resource_path=None, **kwargs):
        super().__init__(*args, **kwargs)
        if resource_path is not None:
            self.resource = "projects/" + CTX.config.APP_NAME + "/databases/(default)/documents/" + resource_path

    def _on_trigger(self, data, context):
        """ For use with Cloud Functions

        :return:
        """
        event_type = context.event_type
        if event_type == EVENT_TYPE_MAPPING['create']:
            snapshot = make_snapshot(data['value'], client=CTX.db)
            self.on_post(snapshot=snapshot)
        elif event_type == EVENT_TYPE_MAPPING['update']:
            before_snapshot = make_snapshot(data['oldValue'], client=CTX.db)
            after_snapshot = make_snapshot(data['value'], client=CTX.db)
            self.on_patch(
                after_snapshot=after_snapshot,
                before_snapshot=before_snapshot)
        elif event_type == EVENT_TYPE_MAPPING['delete']:
            before_snapshot = make_snapshot(data['oldValue'], client=CTX.db)
            self.on_remove(snapshot=before_snapshot)
        elif event_type == EVENT_TYPE_MAPPING['write']:
            before_snapshot = make_snapshot(data['oldValue'], client=CTX.db)
            after_snapshot = make_snapshot(data['value'], client=CTX.db)
            if before_snapshot is None and after_snapshot is not None:
                self.on_post(snapshot=after_snapshot)
            elif before_snapshot is not None and after_snapshot is not None:
                self.on_patch(
                    after_snapshot=after_snapshot,
                    before_snapshot=before_snapshot)
            elif before_snapshot is not None and after_snapshot is None:
                self.on_remove(snapshot=before_snapshot)
            else:
                CTX.logger.error(
                    f"flask-boiler failed to route for: {data} {context}")
                raise ValueError
        else:
            #  TODO: implement mapping for 'write'
            CTX.logger.error(
                msg=f"event_type not supported. {data} {context}")

    def __call__(self, *args, **kwargs):
        try:
            return self._on_trigger(*args, **kwargs)
        except Exception as e:
            CTX.logger.exception(msg="trigger failed")


class OnSnapshotTasksMixin(OnTriggerMixin):
    """
    To be added before ViewModel

    TODO: check contextvar CTX.transaction context switching during
        an await in async function.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.q = Queue()

    def start(self, *args, **kwargs):
        self._start_thread()
        super().start(*args, **kwargs)

    def _main(self):
        """ Push None to self.q to stop the thread

        :return:
        """
        while True:
            item = self.q.get()
            if item is None:
                break
            try:
                item()
            except Exception as e:
                CTX.logger.exception(f"a task in the queue has failed {item}")
            self.q.task_done()

    def _start_thread(self):
        self.thread = Thread(target=self._main, daemon=True)
        self.thread.start()

    def _add_awaitable(self, item):
        """ TODO: test

        :param item:
        :return:
        """
        self.q.put(item)

    def _on_snapshot(self, snapshots, changes, timestamp):
        """
        Note that server reboot will result in some "Modified" objects
            to be routed as "Added" objects

        :param snapshots:
        :param changes:
        :param timestamp:
        """
        for change in changes:
            assert isinstance(change, DocumentChange)
            snapshot = change.document
            assert isinstance(snapshot, DocumentSnapshot)
            try:
                CTX.logger.info(f"DAV: {self.__class__.__name__} started "
                                f"for {snapshot.reference.path}")
                if change.type.name == 'ADDED':
                    self.Protocol.on_create(
                        snapshot=snapshot,
                        mediator=self
                    )
                elif change.type.name == 'MODIFIED':
                    if change.old_index != -1:
                        # Append before snapshot if available
                        before_snapshot = snapshots[change.old_index]
                    else:
                        before_snapshot = None
                    after_snapshot = snapshots[change.new_index]
                    self.on_patch(after_snapshot=after_snapshot,
                                  before_snapshot=before_snapshot)
                elif change.type.name == 'REMOVED':
                    self.Protocol.on_delete(
                        snapshot=snapshot,
                        mediator=self
                    )
            except Exception as e:
                """
                Expects e to be printed to the logger 
                """
                CTX.logger.exception(f"DAV {self.__class__.__name__} failed "
                                     f"for {snapshot.reference.path}")
            else:
                CTX.logger.info(f"DAV {self.__class__.__name__} succeeded "
                                f"or enqueued "
                                f"for {snapshot.reference.path}")

    class Protocol(ProtocolBase):

        @staticmethod
        def on_create(snapshot: DocumentSnapshot, mediator):
            # if snapshot.update_time == snapshot.create_time:
            f = functools.partial(mediator.on_post, snapshot=snapshot)
            # else:
            #     f = functools.partial(mediator.on_patch, after_snapshot=snapshot)
            mediator._add_awaitable(f)

        @staticmethod
        def on_update(snapshot: DocumentSnapshot, mediator):
            f = functools.partial(mediator.on_patch, snapshot=snapshot)
            mediator._add_awaitable(f)

        @staticmethod
        def on_delete(snapshot: DocumentSnapshot, mediator):
            f = functools.partial(mediator.on_remove, snapshot=snapshot)
            mediator._add_awaitable(f)

    def on_post(self, snapshot: DocumentSnapshot):
        """
        Called when a document is 'ADDED' to the result of a query.
        (Will be enqueued to a different thread and run concurrently)

        :param snapshot: Firestore Snapshot added
        """
        pass

    def on_patch(
            self,
            after_snapshot: DocumentSnapshot,
            before_snapshot: Optional[DocumentSnapshot] = None,
    ):
        """ Called when a document is 'MODIFIED' in the result of a query.
        (Will be enqueued to a different thread and run concurrently)

        :param after_snapshot:
        :param before_snapshot: May be None if timestamp set to watcher
            is later than the last_updated time of a snapshot
        :return:
        """
        pass

    def on_remove(self, snapshot: DocumentSnapshot):
        """ Called when a document is 'REMOVED' in the result of a query.
        (Will be enqueued to a different thread and run concurrently)

        :param snapshot: Firestore Snapshot deleted
        """
        pass


def make_snapshot(value, client) -> Optional[DocumentSnapshot]:
    """ Converts a json representation of Document protobuf to DocumentSnapshot
    TODO: note that this method uses _helpers and other internal methods
    TODO: of google cloud library


    :param value:
    :param client:
    :return:
    """

    # Ref: https://developers.google.com/protocol-buffers/docs/pythontutorial
    if value == dict():
        return None

    from google.protobuf.json_format import ParseDict
    from google.cloud.firestore_v1.proto.document_pb2 import Document
    from google.cloud.firestore_v1 import _helpers

    # value is a deserialized (json) form of "Document" protobuf message
    document_pb = ParseDict(value, Document())

    exists = True
    create_time = document_pb.create_time
    update_time = document_pb.update_time
    reference = client.document(document_pb.name)
    data = _helpers.decode_dict(document_pb.fields, client)

    snapshot = DocumentSnapshot(
        reference=reference,
        data=data,
        exists=exists,
        read_time=None,  # No server read_time available
        create_time=create_time,
        update_time=update_time,
    )

    return snapshot


def create_mutation_protocol(mutation):
    class _MutationProtocol:

        @staticmethod
        def on_create(snapshot, mediator):
            mutation.mutate_create(
                doc_id=snapshot.reference.id,
                data=snapshot.to_dict()
            )

    return _MutationProtocol



from onto.common import _NA
from onto.database import Database, Reference, Snapshot
from onto.context import Context as CTX
from couchbase import bucket


class CouchbaseDatabase(Database):

    @classmethod
    def bucket(cls) -> bucket.Bucket:
        return CTX.dbs.couchbase

    @classmethod
    def set(cls, ref: Reference, snapshot: Snapshot, transaction=_NA):
        cls.bucket().collection(collection_name=ref.first).upsert(
            key=ref.last, value=snapshot.to_dict()
        )

    @classmethod
    def create(cls, ref: Reference, snapshot: Snapshot, transaction=_NA):
        cls.bucket().collection(collection_name=ref.first).insert(
            key=ref.last, value=snapshot.to_dict()
        )

    @classmethod
    def delete(cls, ref: Reference, transaction=_NA):
        cls.bucket().collection(collection_name=ref.first).remove(
            key='ref.last'
        )

    @classmethod
    def update(cls, ref: Reference, snapshot: Snapshot, transaction=_NA):
        cls.set(ref, snapshot, transaction)



from onto import schema, fields, domain_model, factory


class UserSchema(schema.Schema):
    followers = fields.Raw()


class UserBase(domain_model.DomainModel):

    class Meta:
        collection_name = "users"


class User(UserBase):

    class Meta:
        schema_cls = UserSchema



from typing import Optional, Dict

from marshmallow.utils import is_iterable_but_not_string

from onto import errors
from onto.mapper import fields
from onto.context import Context as CTX
from onto.mapper.helpers import EmbeddedElement
from ..common import _NA
from ..database import Snapshot
from ..registry import ModelRegistry


def resolve_obj_cls(*, cls, d):
    super_cls, obj_cls = cls, cls

    from onto.common import read_obj_type
    obj_type_str = read_obj_type(d, obj_cls)
    # Note: "obj_type" is NOT the "attribute" property
    # of obj_type field instance

    if obj_type_str is not None:
        """ If obj_type string is specified, use it instead of cls supplied 
            to from_dict. 

        TODO: find a way to raise error when obj_type_str reading 
            fails with None and obj_type evaluates to cls supplied 
            to from_dict unintentionally. 

        """
        obj_cls = ModelRegistry.get_cls_from_name(obj_type_str)
        if obj_cls is None:
            """ If obj_type string is specified but invalid, 
                throw a ValueError. 
            """
            raise ValueError("Cannot read obj_type string: {}. "
                             "Make sure that obj_type is a subclass of {}."
                             .format(obj_type_str, super_cls))

    return obj_cls


class Importable:
    """
    When object subclass this mixin class, the object has the ability to
        be deserialized/loaded/set from a dictionary.
    """

    @classmethod
    def _import_val(cls, val, partial=False, **kwargs):
        """ IMPORTANT! Note that the code may fail and result in
                infinite call chain when there is a circular reference.
                The code is not designed to fail this way, but
                    accidental changes to code may make this behavior
                    possible.

        :param val:
        :param to_get:
        :param kwargs:
        :return:
        """

        def embed_element(val: EmbeddedElement):
            d = val.d
            obj_cls = val.obj_cls
            # data = {key: val for key, val in d.items()}
            obj = obj_cls.from_dict(d=d, partial=partial)
            return obj

        if isinstance(val, EmbeddedElement):
            return embed_element(val)
        elif is_iterable_but_not_string(val):
            if isinstance(val, list):
                val_list = [cls._import_val(elem, partial=partial, **kwargs)
                            for elem in val]
                return val_list
            elif isinstance(val, dict):
                val_d = dict()
                for k, v in val.items():
                    val_d[k] = cls._import_val(v, partial=partial, **kwargs)
                return val_d
            else:
                raise NotImplementedError

        else:
            return val

    def update_vals(self,
                    with_dict: Optional[Dict] = None,
                    with_raw: Optional[Dict] = None,
                    **kwargs) -> None:
        """ Update values of the object in batch.

        :param with_dict:  Update object with keys and values in the
            with_dict.
        :param with_raw: Do field deserialization before setting
            value of attributes.
        :param kwargs:
        """

        if with_dict is None:
            with_dict = dict()
        if with_raw is not None:
            d = self.get_schema_obj().load(with_raw)
            with_dict = {
                key: self._import_val(val)
                for key, val in d.items()
            }
        kwargs_new = {**with_dict, **kwargs}
        for key, val in kwargs_new.items():
            setattr(self, key, val)

    @classmethod
    def _import_from_dict(cls, d, **kwargs):
        return cls._import_val(d, **kwargs)

    @classmethod
    def from_dict(cls, d, transaction=_NA, partial=False, **kwargs):

        obj_cls = resolve_obj_cls(cls=cls, d=d)

        schema_obj = obj_cls.get_schema_obj()
        d = schema_obj.load(d, partial=partial)

        d = cls._import_from_dict(d, partial=partial, transaction=transaction)

        instance = obj_cls.new(**d, **kwargs)  # TODO: fix unexpected arguments
        return instance

    @classmethod
    def from_dict_special(cls, d, **kwargs):
        from onto.models.utils import _collect_attrs

        def y(klass):
            for key, attr in _collect_attrs(klass):
                from onto.attrs.unit import MonadContext
                with MonadContext.context().attribute_name(key).data_key_from_name():
                    if attr.descendant_of(('doc_id',)):
                        yield attr.properties.data_key, attr.name

        special_fields = dict(y(cls))
        m = {
            special_fields[k]: d[k]
            for k, v in d.items()
            if k in special_fields
        }
        d = {
            k:v
            for k, v in d.items()
            if k not in special_fields
        }
        return cls.from_dict(d=d, **m, **kwargs)



class Exportable:
    """
    When object subclass this mixin class, the object has the ability to
        be serialized/dumped/output into a dictionary.
    """

    def _export_val(self, val, **kwargs):
        """
        Private method for serializing an instance variable of the object.
        TODO: test nested

        :param val:
        :param to_save: If set to True, the changes made to an
                    object referenced by master object in
                    a relationship field will be saved. Otherwise,
                    changes are not saved, and the
                    TODO: Add support for atomicity
        """

        def embed_element(val: EmbeddedElement):
            obj = val.obj
            if obj is None:
                return None
            return obj._export_as_dict(**kwargs)  # TODO: remove to_save from elsewhere

        if isinstance(val, EmbeddedElement):
            return embed_element(val)
        elif is_iterable_but_not_string(val):
            if isinstance(val, list):
                val_list = [self._export_val(elem, **kwargs) for elem in val]
                return val_list
            elif isinstance(val, dict):
                val_d = dict()
                for k, v in val.items():
                    val_d[k] = self._export_val(v, **kwargs)
                return val_d
            else:
                raise NotImplementedError
        else:
            return val

    def _export_as_dict(self, **kwargs) -> dict:
        """ Map/dict is only supported at root level for now

        """
        d = self.schema_obj.dump(self)
        return self._export_val(val=d, **kwargs)

    def _export_val_view(self, val):

        def embed_element(val: EmbeddedElement):
            obj = val.obj
            return obj._export_as_view_dict()

        if isinstance(val, EmbeddedElement):
            return embed_element(val)
        elif is_iterable_but_not_string(val):
            if isinstance(val, list):
                val_list = [self._export_val_view(elem) for elem in val]
                return val_list
            elif isinstance(val, dict):
                val_d = dict()
                for k, v in val.items():
                    val_d[k] = self._export_val_view(v)
                return val_d
            else:
                raise NotImplementedError
        else:
            return val

    # def _export_as_view_dict(self) -> dict:
    #     """ Export as dictionary representable to front end.
    #
    #     """
    #     d = self.schema_obj.dump(self)
    #     for key in self.schema_cls._get_reserved_fieldnames():
    #         _ = d.pop(key, None)
    #
    #     return self._export_val_view(d)

    def to_dict(self):
        return self._export_as_dict()

    def to_snapshot(self):
        return Snapshot(**self._export_as_dict())

    # def to_view_dict(self):
    #     return self._export_as_view_dict()


def is_pony(*, klass):
    try:
        from pony.orm.core import Entity
        return issubclass(klass, Entity)
    except ImportError:
        return False


class NewMixin:
    """
    Mixin class for initializing instance variable on creation.
            Needs to be top level.
            Example:
                 SomeClass(..., SomeBase, NewMixin, object):
                    pass

    """

    @classmethod
    def new(cls, allow_default=True, **kwargs):
        """ Instantiates a new instance to the model.

        This is similar to the use of "new" in Java. It is recommended that
        you use "new" to initialize an object, rather than the native
        initializer. Values are initialized based on the order that they are
        declared in the schema.

        :param allow_default: if set to False, an error will be
            raised if value is not provided for a field.
        :param kwargs: keyword arguments to pass to the class
            initializer.
        :return: the instance created
        """

        # fd = cls._get_fields()  # Calls classmethod

        # field_keys = {key for key, field in fd.items() }

        _with_dict = dict()

        for key, field in cls._get_fields().items():
            if field.dump_only:
                continue
            if key in kwargs:
                _with_dict[key] = kwargs[key]
            else:
                if field.required:
                    raise errors.DefaultNotAllowedError
                v = getattr(field, 'default_value', None)
                if v != fields.allow_missing:
                    _with_dict[key] = v

        kwargs_keys = set(kwargs.keys())
        keys_super = kwargs_keys - set(_with_dict.keys())

        d_super = {key: kwargs[key] for key in keys_super}

        return cls(_with_dict=_with_dict, **d_super)

    def _init_(self, *, _with_dict, _kwargs):
        # TODO: note that obj_type and other irrelevant fields are set; fix
        for key, val in _with_dict.items():
            try:
                setattr(self, key, val)
            except AttributeError as ae:
                CTX.logger.error(f"Error encountered while setting key: {key}"
                                 f" and value: {val}."
                                 f" with_dict: {_with_dict}. "
                                 f"Error: {ae.args}")
                raise ae
            # elif isinstance(getattr(self.__class__, key), property):
            #     setattr(self, key, val)
        # TODO: log and throw with extraneous arguments
        if len(_kwargs) != 0:
            raise ValueError(f'Received extraneous arguments '
                             f'{_kwargs}')
        super().__init__(**_kwargs)

    def _init_pony_(self, *, _with_dict, _kwargs):
        if len(_kwargs) != 0:
            raise ValueError(f'Received extraneous arguments '
                             f'{_kwargs}')
        from onto.models.base import PonyStore
        self._attrs = PonyStore()
        self._attrs._set_owner(self)  # TODO: implement garbage collection
        super().__init__(**_with_dict, **_kwargs)

    def __init__(self, _with_dict=None, **kwargs):
        """ Private initializer; do not call directly.
            Use "YourModelClass.new(...)" instead.
        Needs to be top level

        :param _with_dict:
        :param kwargs:
        """
        if _with_dict is None:
            _with_dict = dict()
        if is_pony(klass=self.__class__):
            self._init_pony_(_with_dict=_with_dict, _kwargs=kwargs)
        else:
            self._init_(_with_dict=_with_dict, _kwargs=kwargs)





import abc

from onto.context import Context as CTX
from collections import namedtuple
from typing import Optional, Tuple

# from google.cloud.firestore import DocumentSnapshot, CollectionReference
from onto.mapper.fields import argument, OBJ_TYPE_ATTR_NAME
from . import cmp
import weakref


class QueryBase:
    """
    Query depends on Database
    """

    """
    Abstract property: database 
    must override
    """

    def __init__(self, ref=None, path=None, arguments=None):
        if path is not None:
            from onto.database import Reference
            ref = Reference.from_str(path)
        self.ref = ref
        if arguments is None:
            arguments = list()
        self.arguments = arguments

    @staticmethod
    def _append_original(*args, cur_arguments=None):
        if cur_arguments is None:
            raise ValueError

        if len(args) != 0:
            if len(args) % 3 != 0:
                raise ValueError
            else:
                arg_stack = list(args)

                while len(arg_stack) != 0:
                    arg1 = arg_stack.pop(0)
                    arg2 = arg_stack.pop(0)
                    arg3 = arg_stack.pop(0)
                    cur_arguments.append( argument(
                        key=arg1, comparator=arg2, val=arg3
                    ))

        return cur_arguments

    @staticmethod
    def _append_cmp_style(*args, cur_arguments=None):

        if cur_arguments is None:
            raise ValueError

        if len(args) != 0:

            arg_stack = list(args)

            while len(arg_stack) != 0:

                condition = arg_stack.pop(0)

                key = condition.attr_name

                for comp, other in condition.constraints:
                    if comp == "_in":
                        # Reverse argument order for "in" comparator
                        cur_arguments.append( argument(
                            key=other, comparator="in", val=key)
                        )
                    else:
                        # Append comparator for normal cases
                        cur_arguments.append( argument(
                            key=key, comparator=comp, val=other)
                        )

        return cur_arguments

    def make_copy(self, arguments):
        return self.__class__(ref=self.ref, arguments=arguments)

    def where(self, *args, **kwargs):
        cmp_args = [arg for arg in args if isinstance(arg, cmp.Condition)]
        remaining_args = [arg for arg in args
                          if not isinstance(arg, cmp.Condition)]
        arguments = self.arguments.copy()
        arguments = self._append_cmp_style(
            *cmp_args, cur_arguments=arguments)
        arguments = self._append_original(
            *remaining_args, cur_arguments=arguments)
        return self.make_copy(arguments=arguments)


class Query(QueryBase):
    pass


class ViewModelQuery(QueryBase):

    @classmethod
    def patch_query(cls, parent):
        path = "**/{}_PATCH".format(parent.__name__)
        return cls(parent=parent, path=path)

    @classmethod
    def from_view_model(cls, parent):
        ref = parent._get_collection()
        return cls(parent=parent, ref=ref)

    def __init__(self, parent=None, **kwargs):
        self.parent = parent
        super().__init__(**kwargs)

    def _to_firestore_query(self):
        """ Returns a query with parent=cls._get_collection(), and
                limits to obj_type of subclass of cls.
        """
        from onto.database.firestore import FirestoreDatabase

        db: FirestoreDatabase = CTX.db
        if self.ref.first == '**':
            cur_where = db.firestore_client.collection_group(self.ref.last)
        else:
            q = db._doc_ref_from_ref(self.ref)
            from google.cloud import firestore
            cur_where = firestore.Query(parent=q)
        if len(self.arguments) != 0:
            raise ValueError
        return cur_where


class DomainModelQuery(QueryBase):

    # def get_query(self):
    #     """ Returns a query with parent=cls._get_collection(), and
    #             limits to obj_type of subclass of cls.
    #     """
    #     from onto.database.firestore import FirestoreDatabase
    #     db: FirestoreDatabase = CTX.db
    #     from google.cloud import firestore
    #     cur_where = firestore.Query(
    #         parent=db._doc_ref_from_ref(self.parent()._get_collection())
    #     )
    #     condition = self.parent().get_obj_type_condition()
    #     if condition is not None:
    #         cur_where = cur_where.where(*condition)
    #     return cur_where

    def __init__(self, parent=None, ref=None, **kwargs):
        self.parent = parent
        if ref is None:
            ref = self.parent._get_collection()
        super().__init__(ref=ref, **kwargs)

    def make_copy(self, arguments):
        return self.__class__(
            ref=self.ref, parent=self.parent, arguments=arguments)

    def _to_qualifier(self):
        """ Returns a greedy qualifier. Performance aside, it should work.
        """
        condition = self.parent.get_obj_type_condition()
        arguments = [ condition, *self.arguments ]
        def qualifier(snapshot):
            for (key, comparator, val) in arguments:
                if key not in snapshot:
                    continue
                a = snapshot[key]
                if not comparator(a, val):
                    return False
            return True
        return qualifier


    def _to_firestore_query(self):
        """ Returns a query with parent=cls._get_collection(), and
                limits to obj_type of subclass of cls.
        """
        from onto.database.firestore import FirestoreDatabase

        # TODO: move

        db: FirestoreDatabase = CTX.db
        if self.ref.first == '**':
            cur_where = db.firestore_client.collection_group(self.ref.last)
        else:
            cur_where = db._doc_ref_from_ref(self.ref)
        # cur_where = firestore.Query(parent=q)
        condition = self.parent.get_obj_type_condition()
        if condition is not None:
            arguments = self.arguments + [condition]
        else:
            arguments = self.arguments.copy()
        for (key, comparator, val) in arguments:
            # TODO: NOTE: data_key will always be translated
            data_key = self.parent._query_schema().fields[key].data_key
            condition = comparator if isinstance(comparator, str) else comparator.condition
            # TODO: translate val
            cur_where = cur_where.where(data_key, condition, val)
        return cur_where

    def _to_leancloud_query(self):
        from onto.database.leancloud import LeancloudDatabase

        # db: LeancloudDatabase = CTX.dbs.leancloud  # TODO: read db elsewhere

        condition = self.parent.get_obj_type_condition()

        if condition is not None:
            arguments = self.arguments + [condition]
        else:
            arguments = self.arguments.copy()

        cla_str = self.ref.last
        import leancloud
        cla = leancloud.Object.extend(name=cla_str)
        q = cla.query

        for key, comparator, val in arguments:
            data_key = self.parent._query_schema().fields[key].data_key
            func_name = comparator.condition
            f = getattr(q, func_name)
            f(data_key, val)

        return q



from unittest import mock
from unittest.mock import Mock, patch

import flask
import pytest

from flask import Flask
from onto import auth

import flask_restful
from flask_restful import Resource, ResponseBase
from firebase_admin import auth as firebase_admin_auth
import firebase_admin

from .fixtures import CTX


def get_mock_auth_headers(uid="testuid1"):
    # user = auth.get_user(uid)
    # userIdToken = user.tokens_valid_after_timestamp
    # # userIdToken = auth.create_custom_token(uid=uid, app=firebaseApp)
    # userIdTokenMock = userIdToken
    # warnings.warn("Note that userIdTokenMock is temporary and the test may fail when the token is no longer valid.")
    user_id_token_mock = uid  # For mocking decoding token as uid
    headers = {'Authorization': user_id_token_mock}
    return headers


# @pytest.mark.usefixtures("CTX")
def test_auth(CTX):
    assert CTX.debug

    class IntegerResource(Resource):

        @auth.authenticate
        def get(self, uid):
            assert uid == "test_user_id_2"

    flask_app = flask.app.Flask(__name__)
    api = flask_restful.Api(flask_app)
    api.add_resource(IntegerResource, "/int_resource")
    app = flask_app.test_client()

    app.get(
        path="int_resource/",
        headers=get_mock_auth_headers(uid="test_user_id_2"))


def test_auth_no_headers(CTX):

    assert CTX.debug

    class IntegerResource(Resource):

        @auth.authenticate
        def get(self, uid):
            assert uid == "test_user_id_2"

    flask_app = flask.app.Flask(__name__)
    api = flask_restful.Api(flask_app)
    api.add_resource(IntegerResource, "/int_resource")
    app = flask_app.test_client()

    res = app.get(path="/int_resource")

    assert res.status_code == 401


@pytest.fixture
def vit(monkeypatch):
    # mock_auth = Mock()

    verify_id_token = Mock(
        return_value={
            "uid": "uid_1"
        })
    monkeypatch.setattr(
        firebase_admin_auth,
        "verify_id_token",
        verify_id_token
    )
    return verify_id_token


@pytest.fixture
def MockProductionCTX(CTX, monkeypatch):

    monkeypatch.setattr(
        CTX.config, "TESTING", False
    )


def test_authenticate(MockProductionCTX, vit):

    class IntegerResource(Resource):

        @auth.authenticate
        def get(self, uid):
            assert uid == "uid_1"

    flask_app = flask.app.Flask(__name__)
    api = flask_restful.Api(flask_app)
    api.add_resource(IntegerResource, "/int_resource")
    app = flask_app.test_client()

    res = app.get(
        path="/int_resource",
        headers={
            'Authorization': "correct_id_token_for_uid_1"
        }
    )

    assert vit.call_args[0][0] == "correct_id_token_for_uid_1"


def test_authenticate_testing_config(CTX, vit):
    """ Tests that auth is skipped when CTX.testing is True
    :param CTX:
    :param vit:
    :return:
    """

    class IntegerResource(Resource):

        @auth.authenticate
        def get(self, uid):
            assert uid == "uid_1"

    flask_app = flask.app.Flask(__name__)
    api = flask_restful.Api(flask_app)
    api.add_resource(IntegerResource, "/int_resource")
    app = flask_app.test_client()

    res = app.get(
        path="/int_resource",
        headers={
            'Authorization': "uid_1"
        }
    )

    assert vit.call_count == 0


@pytest.fixture
def vit_bomb(monkeypatch):

    def auth_bomb(*args, **kwargs):
        raise firebase_admin_auth.RevokedIdTokenError("mock message")

    verify_id_token = Mock(wraps=auth_bomb)
    monkeypatch.setattr(
        firebase_admin_auth,
        "verify_id_token",
        verify_id_token
    )
    return verify_id_token


def test_authenticate_error(MockProductionCTX, vit_bomb):

    class IntegerResource(Resource):

        @auth.authenticate
        def get(self, uid):
            assert uid == "uid_1"

    flask_app = flask.app.Flask(__name__)
    api = flask_restful.Api(flask_app)
    api.add_resource(IntegerResource, "/int_resource")
    app = flask_app.test_client()

    res = app.get(
        path="/int_resource",
        headers={
            'Authorization': "correct_id_token_for_uid_1"
        }
    )

    assert res.status_code == 401




import pytest

from onto.attrs import attrs


def test_s():

    from onto.view_model import ViewModel

    from onto.view_model import ViewModelR

    class A(ViewModel):
        # foo = attrs.string
        doc_id = attrs.doc_id

    a = A.new(doc_id='foo')
    assert a.doc_id == 'foo'



"""

Contains test cases inspired by flasgger (MIT licence)
See: https://github.com/flasgger/flasgger/blob/master/LICENSE
"""
from onto import attrs
from flask import Flask
from flasgger import Swagger

from onto.models.base import Serializable
from onto.view import rest_api
from onto.view_model import ViewModel


repo = {

    "doc_id_1": dict(palette_name='cmyk',
                     colors=[
                         {'name': 'cian'},
                         {'name': 'magenta'},
                         {'name': 'yellow'},
                         {'name': 'black'}
                     ], )
}


class ExampleColor(Serializable):
    name = attrs.string()


class ExamplePaletteViewModel(ViewModel):

    class Meta:
        case_conversion = False

    palette_name = attrs.string()
    colors = attrs.embed(obj_cls=ExampleColor, many=True)

    description = "A list of colors (may be filtered by palette)"

    @classmethod
    def get(cls, *args, doc_id, once=True, **kwargs):
        """
        Hard code a palette. (Not standard usage)
        """
        assert once
        obj = super().new(
            *args,
            **repo[doc_id],
            **kwargs
        )
        return obj


# obj = ExamplePaletteViewModel.new(doc_id="doc_id_1")

app = Flask(__name__)
swagger = Swagger(app)

palette_mediator = rest_api.ViewMediator(
    view_model_cls=ExamplePaletteViewModel,
    app=app,
)

palette_mediator.add_instance_get(rule="/palettes/<string:doc_id>")

if __name__ == "__main__":
    """
    Go to http://127.0.0.1:5000/apidocs/ for the auto-generated 
        documentations. 
    """

    from onto.config import Config
    from onto.context import Context as CTX

    if not CTX._ready:
        config = Config(
            app_name="flask-boiler-testing",
            debug=True,
            testing=True,
            certificate_filename="flask-boiler-testing-firebase-adminsdk-4m0ec-7505aaef8d.json"
        )
        CTX.read(config)

    app.run(debug=True)
    # Now visit "http://127.0.0.1:5000/palettes/doc_id_1" to get json


"""
Reserved for testing; Not part of the example 
"""


def test_view_example():
    from onto.config import Config
    from onto.context import Context as CTX

    if not CTX._ready:
        config = Config(
            app_name="flask-boiler-testing",
            debug=True,
            testing=True,
            certificate_filename="flask-boiler-testing-firebase-adminsdk-4m0ec-7505aaef8d.json"
        )
        CTX.read(config)

    assert app.test_client().get(
        "http://127.0.0.1:5000/palettes/doc_id_1"
    ).json == {
               "colors": [
                   {
                       "name": "cian",
                       "obj_type": "dict"
                   },
                   {
                       "name": "magenta",
                       "obj_type": "dict"
                   },
                   {
                       "name": "yellow",
                       "obj_type": "dict"
                   },
                   {
                       "name": "black",
                       "obj_type": "dict"
                   }
               ],
               "palette_name": "cmyk"
           }



import pytest

from examples.fluttergram import domain_models


@pytest.fixture
def users(request, CTX):
    tijuana = domain_models.User.new(doc_id="tijuana")
    tijuana.followers = {"thomasina": True, }
    tijuana.save()

    thomasina = domain_models.User.new(doc_id="thomasina")
    thomasina.followers = dict()
    thomasina.save()

    def fin():
        tijuana.delete()
        thomasina.delete()

    request.addfinalizer(fin)

    return [tijuana, thomasina]


@pytest.fixture
def posts(users, request, CTX):
    p = domain_models.Post.new(doc_id="tj_t", owner_id="tijuana", _remainder={
        "hello": "world"
    })
    p.save()

    def fin():
        p.delete()

    request.addfinalizer(fin)

    return [p]



# import sqlite3
# from flask import current_app, _app_ctx_stack
#
#
# class Boiler(object):
#
#     def __init__(self, app=None):
#         self.app = app
#         if app is not None:
#             self.init_app(app)

    # def init_app(self, app):
    #     app.config.setdefault('SQLITE3_DATABASE', ':memory:')
    #     app.teardown_appcontext(self.teardown)
    #
    # def connect(self):
    #     return sqlite3.connect(current_app.config['SQLITE3_DATABASE'])
    #
    # def teardown(self, exception):
    #     ctx = _app_ctx_stack.top
    #     if hasattr(ctx, 'sqlite3_db'):
    #         ctx.sqlite3_db.close()
    #
    # @property
    # def connection(self):
    #     ctx = _app_ctx_stack.top
    #     if ctx is not None:
    #         if not hasattr(ctx, 'sqlite3_db'):
    #             ctx.sqlite3_db = self.connect()
    #         return ctx.sqlite3_db



import pytest
from google.cloud.firestore_v1 import DocumentSnapshot

from onto.context import Context as CTX
from onto.testing_utils import _wait
from tests.utils import _delete_all

from .views import CityView
from .models import Municipality


@pytest.fixture
def tok_snapshot(request) -> DocumentSnapshot:
    obj = Municipality.new(
        doc_id="TOK", city_name='Tokyo', country='Japan', capital=True)
    obj.save()

    def fin():
        _delete_all(collection_name="City", CTX=CTX)
        _delete_all(collection_name="cityView", CTX=CTX)

    request.addfinalizer(fin)

    return CTX.db.document(obj.doc_ref_str).get()


def test_new_with_snapshot(tok_snapshot):
    view = CityView.new(snapshot=tok_snapshot)


def test_view_mediator(tok_snapshot):
    from .main import city_view_mediator
    city_view_mediator.start()
    _wait()
    res = CTX.db.document("cityView/TOK").get().to_dict()
    assert res == {
        'doc_ref': 'cityView/TOK',
        'obj_type': 'CityView',
        'country': 'Japan',
        'name': 'Tokyo'
    }


@pytest.fixture
def user_form(request):
    doc_ref = CTX.db.document("users/uid1/cityForms/LA")

    def fin():
        _delete_all(subcollection_name="cityForms", CTX=CTX)
        _delete_all(collection_name="City", CTX=CTX)

    request.addfinalizer(fin)

    doc_ref.create(
        dict(
            name='Los Angeles',
            country='USA'
        )
    )


def test_form_mediator(user_form):

    from .main import city_form_mediator
    city_form_mediator.start()
    _wait()
    res = CTX.db.document("City/LA").get().to_dict()

    assert res == {
        'cityName': 'Los Angeles',
        'country': 'USA',
        'doc_id': 'LA',
        'doc_ref': 'City/LA',
        'obj_type': 'City'
    }

    from .main import city_view_mediator
    city_view_mediator.start()
    _wait()
    res = CTX.db.document("cityView/LA").get().to_dict()
    assert res == {
        'name': 'Los Angeles',
        'country': 'USA',
        'doc_ref': 'cityView/LA',
        'obj_type': 'CityView',
    }


"""
Recipe for metaclass __prepare__ 
Plan to use this recipe to pull Attribute objects declared in superclass 
to local namespace to be used directly in class declaration. 

"""

# def test_exp():
#     class Meta(type):
#     def __prepare__(name, bases, **kwds):
#         print("preparing {}".format(name), kwds)
#         my_list = []
#         for b in bases:
#             if hasattr(b, 'my_list'):
#                 my_list.extend(b.my_list)
#         return {'my_list': my_list}
#
# class A(metaclass=Meta):
#     my_list = ["a"]
#
# class B(A):
#     my_list.append(2)
#
# print(A().my_list)
# print(B().my_list)



def test_init():

    from onto.store import Store, reference

    from onto.domain_model import DomainModel
    from onto import attrs

    class U(DomainModel):
        class Meta:
            collection_name = "U"

        t = attrs.bproperty()

    class V(DomainModel):
        class Meta:
            collection_name = "V"

        t = attrs.bproperty()

    class UvStore(Store):

        u = reference(dm_cls=U, many=False)
        v = reference(dm_cls=V, many=False)

    U.new(t='foobar', doc_id='foo').save()
    V.new(t='foobar', doc_id='bar').save()

    s = UvStore.from_dict( {
        'u': (U, 'foo'),
        'v': (V, 'bar')
    }
    )

    print(s)

    struct = dict()
    struct['u'] = (U, 'foo')
    struct['v'] = (V, 'bar')
    store = UvStore.from_struct(struct=struct)

    assert store is not None

    # assert s is None

    u = s.u

    print(u)
    print(s.v)



from examples.meeting_room.tests.test_view_model import *
from examples.meeting_room.tests.test_dav import *
from examples.view_example import test_view_example
from examples.distributed_counter_example import *

from .fixtures import CTX


def test_quickstart_example():
    from importlib import import_module
    _ = import_module("examples.quickstart_example")


def test_relationship_example():
    from importlib import import_module
    _ = import_module("examples.relationship_example")


def test_meeting_quickstart():
    from importlib import import_module
    _ = import_module("examples.meeting_room.quickstart")



"""
See: https://cloud.google.com/functions/docs/deploying/api

gcloud functions deploy to_trigger --runtime python37 --trigger-event providers/cloud.firestore/eventTypes/document.create --trigger-resource "projects/flask-boiler-testing/databases/(default)/documents/gcfTest/{gcfTestDocId}"
"""

from zipfile import ZipFile

def configurations(*, env_vars: dict, resource: str, event_type: str,
                   description: str="", labels: dict,
                   available_memory_mb: int = 256, timeout: str = '540s',
                   entry_point: str = None, name: str, upload_url: str, credentials):
    """
    Docstrings adapted from Google documentations
    Ref:

    :param env_vars:
    :param resource: Required. The resource(s) from which to observe events, for example,
        `projects/_/buckets/myBucket`.

        Not all syntactically correct values are accepted by all services. For
        example:

        1. The authorization model must support it. Google Cloud Functions
           only allows EventTriggers to be deployed that observe resources in the
           same project as the `CloudFunction`.
        2. The resource type must match the pattern expected for an
           `event_type`. For example, an `EventTrigger` that has an
           `event_type` of "google.pubsub.topic.publish" should have a resource
           that matches Google Cloud Pub/Sub topics.

        Additionally, some services may support short names when creating an
        `EventTrigger`. These will always be returned in the normalized "long"
        format.

        See each *service's* documentation for supported formats.
        "service": "A String",
        The hostname of the service that should be observed.

        If no string is provided, the default service implementing the API will
        be used. For example, `storage.googleapis.com` is the default for all
        event types in the `google.storage` namespace.
    :param event_type:
     Required. The type of event to observe. For example:
        `providers/cloud.storage/eventTypes/object.change` and
        `providers/cloud.pubsub/eventTypes/topic.publish`.

        Event types match pattern `providers/*/eventTypes/*.*`.
        The pattern contains:

        1. namespace: For example, `cloud.storage` and
           `google.firebase.analytics`.
        2. resource type: The type of resource on which event occurs. For
           example, the Google Cloud Storage API includes the type `object`.
        3. action: The action that generates the event. For example, action for
           a Google Cloud Storage Object is 'change'.
        These parts are lower case.
    :param description: User-provided description of a function.
    :param labels: Labels associated with this Cloud Function.
    :param available_memory_mb: The amount of memory in MB available for a function.
     Defaults to 256MB.
    :param timeout: The function execution timeout. Execution is considered failed and
    can be terminated if the function is not completed at the end of the
    :param name: A user-defined name of the function. Function names must be unique
     globally and match pattern `projects/*/locations/*/functions/*`
    timeout period. Defaults to 60 seconds. duration must end with 's'.
    :param entry_point: The name of the function (as defined in source code) that will be
    executed. Defaults to the resource name suffix, if not specified. For
    backward compatibility, if function with given name is not found, then the
    system will try to use function named "function".
=        :param upload_url:
    :return:
    """

    if env_vars is None:
        env_vars = dict()

    return {
        "eventTrigger": {
            "resource": resource,
            "eventType": event_type,
        },
        "labels": labels,
        "availableMemoryMb": available_memory_mb,
        "description": description,
        "maxInstances": 42,
        "entryPoint": entry_point,
        "name": name,
        "environmentVariables": env_vars,
        "sourceUploadUrl": upload_url,
        "serviceAccountEmail": credentials.service_account_email,
        "timeout": timeout,
        "ingressSettings": "ALLOW_ALL",
        "runtime": "python37",
    }


def deploy_all(entry_points, env_vars=None):
    """
    Ref: https://cloud.google.com/functions/docs/reference/rest/v1/projects.locations.functions/generateUploadUrl
    """

    if env_vars is None:
        env_vars = dict()

    import googleapiclient.discovery
    from google.oauth2 import service_account
    from onto.config import Config

    config = Config.load()

    SCOPES = ['https://www.googleapis.com/auth/cloud-platform']

    # client = Client.from_service_account_json(config.FIREBASE_CERTIFICATE_JSON_PATH)
    credentials = service_account.Credentials.from_service_account_file(
        config.FIREBASE_CERTIFICATE_JSON_PATH, scopes=SCOPES)

    # Update an existing project
    # project = client.fetch_project(config.APP_NAME)

    # Project name here projects/{project_id}/locations/{location_id}
    project_id = config.APP_NAME
    location_id = "us-central1"
    parent = f"projects/{project_id}/locations/{location_id}"

    functions_service: googleapiclient.discovery.Resource = googleapiclient.discovery.build(
        'cloudfunctions', 'v1', credentials=credentials)
    resp = functions_service.projects().locations().functions().generateUploadUrl(
        parent=parent).execute()

    upload_url = resp['uploadUrl']

    # http://googleapis.github.io/google-api-python-client/docs/dyn/cloudfunctions_v1.projects.locations.functions.html

    import requests

    from git import Repo

    repo = Repo()

    import os

    # Archives current repo and add FIREBASE_CERTIFICATE_JSON_PATH to the zip
    file_path = os.path.join(os.path.curdir, 'repo.zip')
    with open(file_path, 'wb') as fp:
        repo.archive(fp, format='zip')

    # TODO: NOTE that this is supposed to be filename;
    #  this example will fail for all path != name
    # TODO: improve
    cert_json_path = config.FIREBASE_CERTIFICATE_JSON_PATH
    with ZipFile(file_path, 'a') as zipf:
        zipf.write(cert_json_path, cert_json_path)

    with open(file_path, 'rb') as fp:
        resp = requests.put(
            url=upload_url, headers={
                'x-goog-content-length-range': "0,104857600",
                'content-type': 'application/zip',
            }, data=fp
        )

    print(resp)

    print(resp.content)

    if resp.status_code != 200:
        """
        For security reasons, interrupt execution. 
        """
        raise Exception("upload failed")

    # from onto.view.query_delta import OnTriggerMixin

    # entry_points = dict()

    # import os
    #
    # import main
    # print(main)
    #
    # # import importlib.util
    # # import importlib
    # #
    # # main_file_path = os.path.join(os.path.curdir, 'main.py')
    # # spec = importlib.util.spec_from_file_location("main", main_file_path)
    # # main = importlib.util.module_from_spec(spec)
    # # spec.loader.exec_module(main)
    #
    # for var_name in dir(main):
    #     print(var_name)
    #     var = getattr(main, var_name)
    #     print(var)
    #     if issubclass(var.__class__, OnTriggerMixin):
    #         entry_points[var_name] = var
    # else:
    #     print(entry_points)

    for entry_point, mediator in entry_points.items():

        name = f'projects/{project_id}/locations/{location_id}/functions/{entry_point}'

        body = configurations(
            env_vars=env_vars,
            entry_point=entry_point,
            resource=mediator.resource,
            event_type=mediator.TRIGGER_EVENT_TYPE,
            upload_url=upload_url,
            labels=dict(),
            name=name,
            credentials=credentials
        )

        from googleapiclient.errors import HttpError

        try:
            resp = functions_service.projects().locations().functions().get(
                name=name).execute()
            print(resp)
        except HttpError as error:
            if error.resp.status == 404:
                # Create new if no such function existed
                resp = functions_service.projects().locations().functions().create(
                    location=parent, body=body
                ).execute()
                print(resp)
            else:
                raise error
        else:
            # Patch existing functions if one already exists with the same name
            resp = functions_service.projects().locations().functions().patch(
                name=name, body=body
            ).execute()
            print(resp)

        print(resp)


# if __name__ == "__main__":
#     deploy_all()



from onto import domain_model, attrs
# from . import Location, User, Ticket
# from onto.collection_mixin import with_pony, db


class MeetingBase(domain_model.DomainModel
    # , metaclass=with_pony
                  ):
    doc_ref = attrs.doc_ref(type_cls=str)

    class Meta:
        collection_name = "meetings"

    location = attrs.relation(nested=False, dm_cls='Location')
    users = attrs.relation(nested=False, dm_cls='User',
                           collection=list)
    tickets = attrs.relation(nested=False, dm_cls='Ticket',
                             collection=list)
    status = attrs.bproperty(type_cls=str)


class Meeting(MeetingBase,
              # metaclass=with_pony
              ):

    @classmethod
    def _datastore(cls):
        from onto.database.kafka import KafkaDatabase
        return KafkaDatabase


# class MeetingIn(MeetingBase):
#
#     @classmethod
#     def _datastore(cls):
#         from onto.database.kafka import KafkaDatabase
#         return KafkaDatabase




class Identifiable:
    pass


class Entity(Identifiable):
    pass


class ValueObject:
    pass


class Aggregate(Identifiable):

    class Meta:
        is_root = False


class DomainService:
    """
    Stateless
    """
    pass


class WeakReference:
    pass


class StrongReference:
    pass


class ApplicationService:
    pass





from onto import schema, fields, domain_model, view_model
from onto.models.factory import ClsFactory

Schema = schema.Schema


class LuggageItemSchema(Schema):
    luggage_type = fields.Str(data_key="luggage_type")
    weight_in_lbs = fields.Integer(data_key="weight_in_lbs")
    # owner_id = fields.Str(
    #     load_from="owner_id",
    #     dump_to="owner_id",
    #     required=False)


class LuggageCollectionSchema(Schema):
    luggages = fields.Raw(many=True, data_key="luggages")
    total_count = fields.Integer(data_key="total_count", dump_only=True)
    total_weight = fields.Integer(data_key="total_weight", dump_only=True)


class LuggageItemBase(domain_model.DomainModel):

    _collection_name = "LuggageItem"


LuggageItem = ClsFactory.create(
    name="LuggageItem",
    schema=LuggageItemSchema,
    base=LuggageItemBase
)


class LuggagesBase(view_model.ViewModel):
    """ Keeps track of luggage amount
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._luggage_d = dict()

    def get_vm_update_callback(self, dm_cls, *args, **kwargs):
        def update_func(*, vm, dm: dm_cls):
            vm.set(dm.doc_id, dm)
        return update_func

    @property
    def luggages(self):
        res = [self._luggage_d[key] for key in sorted(self._luggage_d)]
        return res

    @luggages.setter
    def luggages(self, val):
        for item in val:
            self._luggage_d[ item.luggage_id ] = item

    @property
    def total_weight(self) -> int:
        return self._get_weight()

    @property
    def total_count(self) -> int:
        return self._get_count()

    def set(self, luggage_id: str, luggage: LuggageItem):
        """ Sets a luggage associated with the luggage_id

        :param luggage_id:
        :param luggage:
        :return:
        """
        self._luggage_d[luggage_id] = luggage

    def _get_count(self) -> int:
        """ Returns the count for luggages by counting all luggages in self._luggage_list.

        :return:
        """
        return len(self._luggage_d)

    def _get_weight(self) -> int:
        """ Returns the weight for luggages by accumulating all luggages in self._luggage_list.

        :return:
        """
        weight = 0

        for key, item in self._luggage_d.items():
            weight += item.weight_in_lbs

        return weight


Luggages = ClsFactory.create(
    name="Luggages",
    schema=LuggageCollectionSchema,
    base=LuggagesBase
)



from onto import fields
from onto.store.store import BPSchema
from onto.context import Context as CTX
from onto.mapper.schema import Schema
from onto.mapper.fields import Integer
from onto.domain_model import DomainModel
from onto.store.struct import Struct
from onto.view.document import ViewMediatorDAV
from onto.view_model import ViewModel
from onto import testing_utils


class ShardSchema(Schema):

    count = Integer()


class Shard(DomainModel):

    class Meta:
        schema_cls = ShardSchema


class ShardsStoreBpss(BPSchema):
    shards = fields.StructuralRef(dm_cls=Shard, many=True)


from onto import attrs

from onto.database import reference


class CounterView(ViewModel):

    total_count = attrs.integer(import_enabled=False)
    shards = attrs.bdict(import_enabled=False, export_enabled=False)
    doc_ref = attrs.doc_ref(import_enabled=False, data_key='doc_ref')

    @doc_ref.getter
    def doc_ref(self):
        return reference/'counters'/'counter_0'
        # return CTX.db.collection("counters").document("counter_0")

    # def __init__(self, *args, **kwargs):
    #     doc_ref =
    #     super().__init__(*args, doc_ref=doc_ref, **kwargs)
    #     self.shards = dict()

    @total_count.getter
    def total_count(self):
        return sum(v.count for _, v in self.store.shards.items())

    def set_shard(self, sid, shard):
        self.shards[sid] = shard

    def get_vm_update_callback(self, dm_cls, *args, **kwargs):

        if dm_cls == Shard:
            def callback(vm: CounterView, dm: Shard):
                vm.set_shard(dm.doc_id, dm)
            return callback
        else:
            return super().get_vm_update_callback(dm_cls, *args, **kwargs)


class CounterMediator(ViewMediatorDAV):

    def __init__(self, shard_size, *args, **kwargs):
        super().__init__(*args, view_model_cls=CounterView, **kwargs)
        self.shard_size = shard_size
        self.view_model = None

    @classmethod
    def notify(cls, obj):
        doc_ref = CTX.db.document(obj.doc_ref)
        doc_ref.set(obj.to_dict())

    def start(self):

        struct = Struct(schema_obj=ShardsStoreBpss())

        for i in range(self.shard_size):
            doc_id = str(i)
            shard = Shard.new(doc_id=doc_id, )
            shard.save()
            struct["shards"][shard.doc_id] = (Shard, doc_id)

        self.view_model = self.view_model_cls.get(
            f_notify=self.notify,
            struct_d=struct,
            once=False)

"""
Reserved for testing 
"""


def test_counter():

    mediator = CounterMediator(shard_size=10)
    mediator.start()

    testing_utils._wait()

    ref = CTX.db.collection("counters").document("counter_0")

    assert ref.get().to_dict()["totalCount"] == 0


def test_increment(CTX):
    """
    TODO: add teardown steps to delete documents generated in firestore

    :param CTX:
    :return:
    """

    mediator = CounterMediator(
        mutation_cls=None,
        shard_size=2
    )

    mediator.start()

    testing_utils._wait(factor=.7)

    doc_ref = CTX.db.collection("counters").document("counter_0")
    snapshot = doc_ref.get()
    assert snapshot.to_dict() == {
        "doc_ref": "counters/counter_0",
        "obj_type": "CounterView",
        "totalCount": 0
    }

    from google.cloud.firestore import Increment

    CTX.db.collection("Shard").document("0").set(
        document_data={
            "count": Increment(1),
        },
        merge=True
    )

    testing_utils._wait()

    assert doc_ref.get().to_dict() == {
        "doc_ref": "counters/counter_0",
        "obj_type": "CounterView",
        "totalCount": 1
    }



import functools

from onto.common import _NA
from onto.database import Database, Reference, Snapshot, Listener
from onto.context import Context as CTX

from google.cloud import firestore
from google.cloud.firestore_v1.document import _get_document_path, \
    DocumentReference, DocumentSnapshot

# TODO: NOTE maximum of 1 firestore client allowed since we used a global var.
from typing import List
from onto.query.query import Query
from onto.store.snapshot_container import SnapshotContainer
from math import inf


FirestoreListener = None

class FirestoreReference(Reference):

    def is_collection(self):
        return len(self.params) % 2 == 1

    @property
    def collection(self):
        return self.first

    def is_document(self):
        return len(self.params) % 2 == 0

    def is_collection_group(self):
        return self.first == "**" and len(self.params) == 2

    @classmethod
    def from__document_name(cls, document_name: str):
        return cls.from_str(document_name)

    @classmethod
    def from_document_reference(cls, ref: DocumentReference):
        return cls.from_str(ref.path)

    @property
    def _document_path(self):
        return str(self)


class FirestoreDatabase(Database):

    firestore_client = None

    @classmethod
    def listener(cls):
        return FirestoreListener


    class Comparators(Database.Comparators):

        eq = '=='
        gt = '>'
        ge = '>='
        lt = '<'
        le = '<='
        contains = 'array_contains'
        _in = 'in'

    @classmethod
    def make_document_path(cls, ref: FirestoreReference):
        """Create and cache the full path for this document.

        Migrated from google.cloud.DocumentReference.document:
            - changed to read document name from ref.path or str(ref)

        Of the form:

            ``projects/{project_id}/databases/{database_id}/...
                  documents/{document_path}``

        Returns:
            str: The full document path.

        Raises:
            ValueError: If the current document reference has no ``client``.
        """
        _document_path_internal = _get_document_path(
            cls.firestore_client,
            ref.path
        )
        return _document_path_internal

    @classmethod
    def transaction(cls):
        return cls.firestore_client.transaction()

    @classmethod
    def _doc_ref_from_ref(cls, ref):
        if ref._is_empty:
            # TODO: change this behavior
            return cls.firestore_client
        elif ref.is_collection_group():
            return cls.firestore_client.collection_group(ref.last)
        elif ref.is_collection():
            return cls.firestore_client.collection(str(ref))
        elif ref.is_document():
            return cls.firestore_client.document(str(ref))
        else:
            raise ValueError

    @classmethod
    def set(cls, ref: Reference, snapshot: Snapshot, transaction=_NA, **kwargs):

        if transaction is _NA:
            transaction = CTX.transaction_var.get()

        doc_ref = cls._doc_ref_from_ref(ref)
        # doc_ref.set(
        #     document_data=snapshot.to_dict()
        # )

        if transaction is None:
            doc_ref.set(document_data=snapshot.to_dict(), **kwargs)
        else:
            transaction.set(reference=doc_ref,
                            document_data=snapshot.to_dict(), **kwargs)

    @classmethod
    def get(cls, ref: Reference, transaction=_NA):
        if transaction is _NA:
            transaction = CTX.transaction_var.get()

        doc_ref = cls._doc_ref_from_ref(ref)

        if transaction is None:
            document_snapshot = doc_ref.get()
        else:
            document_snapshot = doc_ref.get(transaction=transaction)

        return FirestoreSnapshot.from_document_snapshot(
            document_snapshot=document_snapshot)

    @classmethod
    def get_many(cls, refs: [Reference], transaction=_NA):
        if transaction is _NA:
            transaction = CTX.transaction_var.get()

        doc_refs = [cls._doc_ref_from_ref(ref) for ref in refs]

        if transaction is None:
            document_snapshots = cls.firestore_client.get_all(doc_refs)
        else:
            document_snapshots = cls.firestore_client.get_all(
                doc_refs, transaction=transaction)

        for document_snapshot in document_snapshots:
            yield FirestoreReference.from_document_reference(document_snapshot.reference),\
                  FirestoreSnapshot.from_document_snapshot(
                document_snapshot=document_snapshot)

    update = set
    create = set

    @classmethod
    def delete(cls, ref: Reference, transaction=_NA):
        doc_ref = cls._doc_ref_from_ref(ref)
        if transaction is _NA:
            transaction = CTX.transaction_var.get()

        if transaction is None:
            doc_ref.delete()
        else:
            transaction.delete(reference=doc_ref)

    @classmethod
    def query(cls, q: Query):
        for document in q._to_firestore_query().stream():
            assert isinstance(document, DocumentSnapshot)
            ref = FirestoreReference.from_document_reference(document.reference)
            snapshot = FirestoreSnapshot.from_document_snapshot(document)
            yield (ref, snapshot)

    ref = FirestoreReference()


class FirestoreSnapshot(Snapshot):

    @classmethod
    def from_document_snapshot(
            cls, document_snapshot: firestore.DocumentSnapshot):
        return cls(
            document_snapshot.to_dict()
        )

    @classmethod
    def from_data_and_meta(
            cls, **kwargs):
        """ Usage: snapshot = FirestoreSnapshot.from_data_and_meta(
                    # reference=document_ref,
                    data=data,
                    exists=True,
                    read_time=None,
                    create_time=document.create_time,
                    update_time=document.update_time,
                )

        :param kwargs:
        :return:
        """
        DATA_KEYWORD = 'data'
        if DATA_KEYWORD not in kwargs:
            raise ValueError
        else:
            data = kwargs[DATA_KEYWORD]
            __onto_meta__ = {
                key: val
                for key, val in kwargs.items()
                if key != DATA_KEYWORD
            }
            return cls(**data, __onto_meta__=__onto_meta__)

    @classmethod
    def empty(cls, **kwargs):
        return cls.from_data_and_meta(
            data=dict(),
            exists=False,
            **kwargs
        )


from threading import Lock, Condition
import heapq


TARGET_ID_RANGE = (32, 64)


class TargetIdAssigner:
    """
    For now, we allow 32 concurrent targets to test that overflow is handled.
    """

    def __init__(self):
        self.lock = Lock()
        self.sold_out = Condition()
        with self.lock, self.sold_out:
            self._heap = list(range(*TARGET_ID_RANGE))
            heapq.heapify(self._heap)

    def assign_id(self):
        with self.lock, self.sold_out:
            while not len(self._heap) > 0:
                self.sold_out.wait()
            item = heapq.heappop(self._heap)
            return item

    def release_id(self, _id):
        with self.lock, self.sold_out:
            heapq.heappush(self._heap, _id)
            self.sold_out.notify(n=1)


_LOGGER = CTX.logger


from _collections import defaultdict


"""
5 changes:
- Just tracked (previously may exist or not have existed in the datastore)
- Just created (previously did not exist in the datastore) 
- Just removed (conditions does not meet query, but was tracked)
- Just deleted (previously tracked and now deleted)
- Just changed (previously tracked and now updated) 

Each listener acts like an airport surveillance radar.
Each aircraft is like a snapshot with its states. 
 
When the flight enters the area, it is first observed by the listener 
    1. the flight may have departed a long time ago and was just observed when 
    passing through the area ("just tracked"), 
        - State of last snapshot is unknown 
    2. or it could have just taken off, and we need to register the flight 
    as a recently departed flight ("just created")
        - State of last snapshot does not exist 
When the flight is no longer observed by the radar, 
    1. the flight may have just landed, in which case, we need to register 
    the flight as landed ("just deleted")
        - New state does not exist 
    2. the flight may have left the area of observation ("just removed")
        - New state is unknown 
When the flight was already observed by the radar (made initial appearance), 
    1. the flight has changed altitude, in which case, we want to run checking 
    to see if the altitude is safe ("just changed")
        - Previous and new state exist 
    
Depending on the time that the radar was turned on, we may observe the same 
signal as "just created" or "just tracked". We are able to see if the flight 
was just created by checking its create and update timestamp. 

"""


def timestamp_key(t) -> tuple:
    return (t.seconds, t.nanos)


class FirestoreListener(Listener):

    _watch = None
    _containers = defaultdict(SnapshotContainer)

    @classmethod
    def _get_watch(cls):

        from onto.context import Context as CTX
        from onto.watch import _Watch

        cls._watch = _Watch(
            # document_reference=None,
            firestore=CTX.db.firestore_client,
            comparator=lambda d1, d2: 1,
            document_snapshot_cls=DocumentSnapshot,
            document_reference_cls=DocumentReference,
        )

        while not cls._watch._rpc.is_active:
            # TODO: change; This is a temporary impl; wait may never stop
            import time
            time.sleep(0.020)

        return cls._watch

    _assigner = TargetIdAssigner()

    @classmethod
    def register(cls, query, source):
        def callback(*args, **kwargs):
            f = functools.partial(source._call, *args, **kwargs)
            cls._coordinator._add_awaitable(f)
        target_id = cls.for_query(query=query, cb=callback)
        cls._registry[target_id] = source

    @classmethod
    def deregister(cls):
        raise NotImplementedError

    @classmethod
    def _process_proto(cls, target_id, proto):
        """

        This method is ported from from google.cloud.firestore_v1.watch

        Copy of the license of google.cloud.firestore_v1.watch:
        # Copyright 2017 Google LLC All rights reserved.
        #
        # Licensed under the Apache License, Version 2.0 (the "License");
        # you may not use this file except in compliance with the License.
        # You may obtain a copy of the License at
        #
        #     http://www.apache.org/licenses/LICENSE-2.0
        #
        # Unless required by applicable law or agreed to in writing, software
        # distributed under the License is distributed on an "AS IS" BASIS,
        # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
        # See the License for the specific language governing permissions and
        # limitations under the License.

        Difference from source code: the way document_change is stored

        :param proto:
        :param target_id:
        :return:
        """

        container = cls._containers[target_id]

        _firestore = CTX.db.firestore_client

        document_change = getattr(proto, "document_change", "")
        document_delete = getattr(proto, "document_delete", "")
        # document_remove = getattr(proto, "document_remove", "")
        # filter_ = getattr(proto, "filter", "")

        if str(document_change):
            _LOGGER.debug("on_snapshot: document change")

            # No other target_ids can show up here, but we still need to see
            # if the targetId was in the added list or removed list.
            target_ids = document_change.target_ids or []
            removed_target_ids = document_change.removed_target_ids or []
            changed = False
            removed = False

            if target_id in target_ids:
                changed = True

            if target_id in removed_target_ids:
                removed = True

            if changed:
                _LOGGER.debug("on_snapshot: document change: CHANGED")

                # google.cloud.firestore_v1.types.Document
                document = document_change.document
                # TODO: change
                from google.cloud.firestore_v1 import _helpers
                data = _helpers.decode_dict(document.fields, _firestore)

                # Create a snapshot. As Document and Query objects can be
                # passed we need to get a Document Reference in a more manual
                # fashion than self._document_reference
                document_name = document.name

                document_name = cls._trim_document_name(_firestore,
                                                       document_name)

                ref = FirestoreReference.from__document_name(document_name)

                snapshot = FirestoreSnapshot.from_data_and_meta(
                    # reference=document_ref,
                    data=data,
                    exists=True,
                    read_time=None,
                    create_time=document.create_time,
                    update_time=document.update_time,
                )

                # TODO: apply lock
                if not container.has_previous(ref):
                    container.set(
                        key=ref,
                        val=FirestoreSnapshot.empty(
                            create_time=-inf,
                            update_time=-inf,
                            read_time=-inf
                        ),
                        timestamp=(-inf, -inf)
                    )

                prev = container.previous(ref)
                prev.next = snapshot
                snapshot.prev = prev

                container.set(
                    key=ref,
                    val=snapshot,
                    timestamp=timestamp_key(document.update_time)
                )

            elif removed:
                raise NotImplementedError

        elif str(document_delete):
            _LOGGER.debug("on_snapshot: document change: DELETE")
            name = document_delete.document  # should be a str

            document_name = cls._trim_document_name(_firestore, name)
            ref = FirestoreReference.from__document_name(document_name)

            snapshot = FirestoreSnapshot.empty(
                    create_time=None,
                    update_time=None,
                    read_time=document_delete.read_time
            )

            prev = container.previous(ref)
            prev.next = snapshot
            snapshot.prev = prev

            container.set(
                key=ref,
                val=snapshot,
                timestamp=timestamp_key(document_delete.read_time)
            )
        else:
            raise ValueError  # TODO: implement

    @classmethod
    def _trim_document_name(cls, _firestore, document_name):
        db_str = _firestore._database_string
        db_str_documents = db_str + "/documents/"
        if document_name.startswith(db_str_documents):
            document_name = document_name[len(db_str_documents):]
        return document_name

    @classmethod
    def callback(cls, target_id, protos, read_time, *, cb):
        container = cls._containers[target_id]
        with container.lock:
            for proto in protos:
                cls._process_proto(target_id, proto)
            container._read_times.append(
                (read_time.seconds, read_time.nanos)
            )
            cb(container)

    @classmethod
    def for_query(cls, query: Query, cb):
        query = query._to_firestore_query()
        parent_path, _ = query._parent._parent_info()
        from google.cloud.firestore_v1.proto import firestore_pb2
        query_target = firestore_pb2.Target.QueryTarget(
            parent=parent_path, structured_query=query._to_protobuf()
        )

        target_id = cls._assigner.assign_id()
        target = {
            "query": query_target,
            "target_id": target_id
        }
        import functools
        cls._get_watch().add_target(
            target, functools.partial(cls.callback, cb=cb)
        )

        return target_id

    @classmethod
    def for_refs(cls, refs: List[FirestoreReference], cb):
        documents = [FirestoreDatabase.make_document_path(ref) for ref in refs]
        target_id = cls._assigner.assign_id()
        target = {
            "documents": {"documents": documents},
            "target_id": target_id,
        }
        import functools
        cls._get_watch().add_target(
            target, functools.partial(cls.callback, cb=cb)
        )
        return target_id

    @classmethod
    def release_target(cls, target_id):
        cls._watch.remove_target(target_id)



from examples.meeting_room.views.graphql_view import UserGraphQLMediator, UserLoginMediator, UserAddMediator
from onto.sink.graphql import op_schema

# PYTEST FIXTURES: RETAIN
from tests.fixtures import CTX
from examples.meeting_room.tests.fixtures import users, meeting, location, tickets

from tests.test_kafka import kafka_server, zookeeper_proc, kafka_consumer


def test_run(users):
    schema_all = list()

    schema_all += UserGraphQLMediator.start()
    schema_all += UserAddMediator.start()
    # schema_all += UserLoginMediator.start()
    #
    from graphql import GraphQLSchema

    schema = GraphQLSchema(
        query=op_schema(op_type='Query', schema_all=schema_all),
        subscription=op_schema(op_type='Subscription', schema_all=schema_all),
        mutation=op_schema(op_type='Mutation', schema_all=schema_all)
    )

    from stargql import GraphQL

    app = GraphQL(
        schema=schema,
    )

    import asyncio
    loop = asyncio.get_event_loop()

    from uvicorn import Config
    config = Config(app=app, loop=loop, port=8081, debug=True)
    from uvicorn import Server
    server = Server(config)

    # from kafka import KafkaProducer
    # kafka_server = KafkaProducer(bootstrap_servers='10.10.8.140:9092')
    # kafka_server.send('users', value='hello world!'.encode('utf-8'))

    loop.run_until_complete(server.serve())

    # uvicorn.run(app, port=8080, debug=True)



import random
import string
from typing import TypeVar


# Generate a random string
# with 32 characters.
# https://www.geeksforgeeks.org/generating-random-ids-python/
from functools import partial

from inflection import camelize, underscore

from onto.common import _NA
from onto.context import Context as CTX
from onto.database import Reference
from onto.registry import ModelRegistry


def random_id(n=32):
    random_id_str = ''.join([random.choice(string.ascii_letters + string.digits) for n in range(n)])
    return random_id_str


def obj_type_serialize(obj: object):
    """ Returns class name. To be used with Schema.obj_type field.

    :param obj:
    :return:
    """

    return obj.__class__.__name__


def obj_type_deserialize(value):
    """ Returns class name. To be used with Schema.obj_type field.

    :param value:
    :return:
    """

    return value


def attr_name_to_firestore_key(s):
    res = camelize(s, uppercase_first_letter=False)
    if firestore_key_to_attr_name(res) != s:
        raise ValueError("attr_name: {} is not invertible. "
                         .format(s)
                         )
    else:
        return res


camel = attr_name_to_firestore_key


firestore_key_to_attr_name = underscore


T = TypeVar('T', covariant=True)


def snapshot_to_obj(
        snapshot: 'google.cloud.firestore.DocumentSnapshot', reference: Reference,
        super_cls: T = None, **kwargs) -> T:
    """ Converts a firestore document snapshot to FirestoreObject

    :param snapshot: firestore document snapshot
    :param super_cls: subclass of FirestoreObject
    :return:
    """

    # if not snapshot.exists:
    #     return None

    d = snapshot.to_dict()
    obj_cls = super_cls

    if "obj_type" in d:
        obj_type = d["obj_type"]
        obj_cls = ModelRegistry.get_cls_from_name(obj_type)

        if obj_cls is None:
            raise ValueError("Cannot read obj_type: {}. "
                             "Make sure that obj_type is a subclass of {}. "
                             .format(obj_type, super_cls))

    obj = obj_cls.from_dict(d=d, doc_ref=reference, **kwargs)
    return obj


def auto_property(attr_name, inner_attr):
    """ Gets a property object that projects operations on
            self.some_attr_name to self.inner_attr.attr_name

    :param attr_name:
    :param inner_attr:
    :return:
    """
    def fget(self):
        inner = getattr(self, inner_attr)
        return getattr(inner, attr_name)

    def fset(self, value):
        inner = getattr(self, inner_attr)
        setattr(inner, attr_name, value)
    #
    # def fdel(self):
    #     inner = getattr(self, inner_attr)
    #     delattr(inner, attr_name)

    return property(fget=fget, fset=fset)


def doc_ref_from_str(doc_ref_str):
    return CTX.db.ref / doc_ref_str



import weakref

from onto.source.protocol import Protocol


class SourceBase:

    def __set_name__(self, owner, name):
        """ Keeps mediator as a weakref to protect garbage collection
        Note: mediator may be destructed if not maintained or referenced
            by another variable. Mediator needs to be explicitly kept alive.

        :param owner:
        :param name:
        :return:
        """
        self.parent = weakref.ref(owner)


class Source(SourceBase):

    _protocol_cls = Protocol

    def __init__(self) -> object:
        """ Initializes a ViewMediator to declare protocols that
                are called when the results of a query change. Note that
                mediator.start must be called later.

        :param query: a listener will be attached to this query
        """
        self.protocol = self._protocol_cls()

    @property
    def triggers(self):
        return self.protocol

    @property
    def mediator_instance(self):
        return self.parent()()

    def _invoke_mediator(self, *args, func_name, **kwargs):
        fname = self.protocol.fname_of(func_name)
        # TODO: fix: listener close invokes unintended on_delete
        if fname is None:
            raise ValueError(
                f"fail to locate {func_name}"
                f" for {self.mediator_instance.__class__.__name__}"
            )
        f = getattr(self.mediator_instance, fname)
        f(*args, **kwargs)

    async def _invoke_mediator_async(self, *args, func_name, **kwargs):
        fname = self.protocol.fname_of(func_name)
        # TODO: fix: listener close invokes unintended on_delete
        if fname is None:
            raise ValueError(
                f"fail to locate {func_name}"
                f" for {self.mediator_instance.__class__.__name__}"
            )
        f = getattr(self.mediator_instance, fname)
        return await f(*args, **kwargs)



from onto.sink.base import Sink


class JsonRpcSink(Sink):

    def __init__(self, *args, uri, **kwargs):
        self.uri = uri
        super().__init__(*args, **kwargs)

    def emit(self, method, **kwargs):
        from jsonrpcclient import request
        response = request(self.uri, method, **kwargs)
        return response

sink = JsonRpcSink




import warnings

# from google.cloud.firestore_v1 import Transaction

from onto.collection_mixin import CollectionMixin, CollectionMemberMeta
from onto.firestore_object import FirestoreObject
from onto.query.query_mixin import QueryMixin
from onto.models.meta import SerializableMeta
from onto.utils import random_id, doc_ref_from_str

from onto.mapper import schema, fields


class PrimaryObjectSchema(schema.Schema):

    doc_id = fields.DocIdField(
        attribute="doc_id",
        dump_only=True,
        data_key="doc_id",
        required=False
    )

    doc_ref = fields.DocRefField(
        attribute="doc_ref",
        dump_only=True,
        data_key="doc_ref",
        required=False
    )


def _collect_query_schema(klass):
    d = dict(klass.get_schema_obj().fields)
    for child in klass._get_children():
        for key, val in child.get_schema_obj().fields.items():
            field_obj = val

            if key in d and d[key] != field_obj:
                warnings.warn(
                    f"Conflict when resolving field for {key}. The "
                    f"field is for querying database "
                    f"from a parent PrimaryObject when "
                    f"such base has no declared schema. ")
            else:
                d[key] = field_obj
    tmp_schema = klass._schema_base.from_dict(d)
    return tmp_schema


class PrimaryObject(FirestoreObject, QueryMixin, CollectionMixin,
                    metaclass=CollectionMemberMeta):
    """
    Primary Object is placed in a collection in root directory only.
    the document will be stored in and accessed from
            self.collection.document(doc_id)

    """

    _collection_name = None
    _schema_base = PrimaryObjectSchema

    @classmethod
    def _query_schema(cls):
        return _collect_query_schema(cls)()

    # @classmethod
    # def get_schema_cls(cls):
    #     """ Returns schema_cls or the union of all schemas of subclasses.
    #             Should only be used on the root DomainModel. Does not
    #             cache the result.
    #
    #     """
    #     d = dict()
    #     if super().get_schema_cls() is None:
    #         for child in cls._get_children():
    #             for key, val in child.get_schema_obj().fields.items():
    #                 field_obj = val
    #
    #                 if key in d and d[key] != field_obj:
    #                     warnings.warn(
    #                         f"Conflict when resolving field for {key}. The "
    #                         f"field is for querying database "
    #                         f"from a parent PrimaryObject when "
    #                         f"such base has no declared schema. ")
    #                 else:
    #                     d[key] = field_obj
    #         tmp_schema = cls._schema_base.from_dict(d)
    #         return tmp_schema
    #     else:
    #         return super().get_schema_cls()

    @property
    def doc_id(self):
        """ Returns Document ID
        """
        return self.doc_ref.id
    #
    # @property
    # def doc_ref(self):
    #     """ Returns Document Reference
    #     """
    #     return self._doc_ref

    random_id = random_id

    @classmethod
    def new(cls, doc_id=None, doc_ref=None, **kwargs):
        """ Creates an instance of object and assign a firestore reference
        with random id to the instance. This is similar to the use
        of "new" in Java. It is recommended that you use "new" to
        initialize an object, rather than the native initializer.
        Values are initialized based on the order that they are
        declared in the schema.

        :param: doc_id: Document ID
        :param: doc_ref: Document Reference
        :param allow_default: if set to False, an error will be
            raised if value is not provided for a field.
        :param kwargs: keyword arguments to pass to the class
            initializer.
        :return: the instance created
        """
        if doc_ref is None:
            if doc_id is None:
                doc_id = cls.random_id()
            doc_ref = cls.ref_from_id(doc_id=doc_id)
        obj = super().new(doc_ref=doc_ref, **kwargs)
        return obj

    @classmethod
    def get(cls, *, doc_ref_str=None, doc_ref=None, doc_id=None,
            transaction: 'google.cloud.firestore_v1.Transaction'=None):
        """ Returns the instance from doc_id.

        :param doc_ref_str: DocumentReference path string
        :param doc_ref: DocumentReference
        :param doc_id: gets the instance from self.collection.document(doc_id)
        :param transaction: firestore transaction
        """

        if doc_ref_str is not None:
            doc_ref = doc_ref_from_str(doc_ref_str)

        if doc_ref is None:
            doc_ref = cls.ref_from_id(doc_id=doc_id)

        return super().get(doc_ref=doc_ref, transaction=transaction)



from .models import City
from onto.context import Context as CTX
from onto.view_model import ViewModel, Struct
from onto.store import Store, reference
from onto import attrs


class CityForm(ViewModel):

    name = attrs.bproperty()
    country = attrs.bproperty()
    city_id = attrs.bproperty()

    city = attrs.bproperty(initialize=True)

    @city.init
    def city(self):
        self._attrs.city = City.new(doc_id=self.doc_ref.id)

    @name.setter
    def name(self, val):
        self.city.city_name = val

    @country.setter
    def country(self, val):
        self.city.country = val

    def propagate_change(self):
        self.city.save()



import contextlib
import contextvars

from inflection import camelize


class ContextVariable(contextlib.ContextDecorator):
    stack = list()
    var: contextvars.ContextVar = None

    def __init__(self, next_var):
        self.next_var = next_var

    def __enter__(self):
        # TODO: not atomic when multithreading
        self.stack.append(self.var.set(self.next_var))

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.var.reset(self.stack.pop())
        return False

    @classmethod
    def get(cls):
        return cls.var.get()


def make_variable(name, **kwargs):
    class InlineContextVariable(ContextVariable):
        var = contextvars.ContextVar(name, **kwargs)
    return InlineContextVariable


def unpack_dict(d):
    import inspect
    from onto.helpers.unpack_tuple_ported import get_assigned_name
    names = get_assigned_name(inspect.currentframe().f_back)
    for name in names:
        yield d[name]
    yield from ()


def unpack_dict_decamelize(d):
    import inspect
    from onto.helpers.unpack_tuple_ported import get_assigned_name
    names = get_assigned_name(inspect.currentframe().f_back)
    for name in names:
        yield d[camelize(name, uppercase_first_letter=False)]
    yield from ()



from examples.meeting_room.domain_models import Meeting
from examples.meeting_room.view_models import MeetingSession
from onto.query.query import ViewModelQuery
from onto.view import Mediator


class MeetingSessionPatch(Mediator):
    from onto import source, sink

    source = source.firestore(
        query=ViewModelQuery.patch_query(parent=MeetingSession)
    )
    sink = sink.firestore()

    @source.triggers.on_create
    def patch_meeting(self, ref, snapshot):
        """ TODO: add transaction

        :param ref:
        :param snapshot:
        :return:
        """
        d = snapshot.to_dict()
        meeting_id = d["target_meeting_id"]

        d = {
            Meeting.get_schema_cls().g(key): val
            for key, val in d.items() if key in {"inSession",}
        }

        user_id = ref.params[1]

        obj = MeetingSession.get(
            doc_id=meeting_id,
            once=True,
        )
        obj.update_vals(user_id=user_id, with_dict=d)
        # TODO: switch to notify
        obj.propagate_change()


class MeetingSessionGet(Mediator):

    from onto import source, sink

    source = source.domain_model(Meeting)
    sink = sink.firestore()  # TODO: check variable resolution order

    @source.triggers.on_update
    @source.triggers.on_create
    def materialize_meeting_session(self, obj):
        meeting = obj
        assert isinstance(meeting, Meeting)

        def notify(obj):
            for ref in obj._view_refs:
                self.sink.emit(reference=ref, snapshot=obj.to_snapshot())

        _ = MeetingSession.get(
            doc_id=meeting.doc_id,
            once=False,
            f_notify=notify
        )
        # mediator.notify(obj=obj)

    @classmethod
    def start(cls):
        cls.source.start()


class MeetingSessionLc(MeetingSession):

    @classmethod
    def _datastore(cls):
        from onto.context import Context as CTX
        return CTX.dbs.leancloud


class MeetingSessionShow(Mediator):

    from onto import source
    from onto.sink.leancloud import LcSink

    source = source.domain_model(Meeting)
    sink = LcSink()  # TODO: check variable resolution order

    @source.triggers.on_update
    @source.triggers.on_create
    def materialize_meeting_session(self, obj):
        meeting = obj
        assert isinstance(meeting, Meeting)

        def notify(obj):
            for ref in obj._view_refs:
                self.sink.emit(ref=ref, snapshot=obj.to_snapshot())

        _ = MeetingSessionLc.get(
            doc_id=meeting.doc_id,
            once=False,
            f_notify=notify
        )
        # mediator.notify(obj=obj)

    @classmethod
    def start(cls):
        cls.source.start()



"""
Some examples are inspired by firestore documentations, some copyright
    conditions of the firestore documentations apply to code on this page.

    https://firebase.google.com/docs/firestore/query-data/queries
"""

import pytest
from google.cloud.firestore import Query, DocumentReference, \
    CollectionReference

from onto.config import Config
from onto.context import Context as CTX

from onto.domain_model import DomainModel
from onto import attrs

# For pytest, DO NOT DELETE
from .fixtures import *

from .utils import _delete_all


class CityM(DomainModel):

    @classmethod
    def _datastore(cls):
        from onto.context import Context as CTX
        return CTX.dbs.leancloud

    city_name = attrs.bproperty()
    country = attrs.bproperty()
    capital = attrs.bproperty()

    class Meta:
        collection_name = "CityM"


class MunicipalityM(CityM):
    pass


class StandardCityM(CityM):
    city_state = attrs.bproperty()
    regions = attrs.bproperty()


@pytest.fixture
def setup_cities(request, CTX):

    def fin():
        for obj in CityM.all():
            obj.delete()

    request.addfinalizer(fin)

    sf = StandardCityM.new(doc_id="SF")

    sf.city_name, sf.city_state, sf.country, sf.capital, sf.regions = \
        'San Francisco', 'CA', 'USA', False, ['west_coast', 'norcal']
    sf.save()

    la = StandardCityM.new(doc_id="LA")
    la.city_name, la.city_state, la.country, la.capital, la.regions = \
        'Los Angeles', 'CA', 'USA', False, ['west_coast', 'socal']
    la.save()

    dc = MunicipalityM.new(doc_id="DC")
    dc.city_name, dc.country, dc.capital = 'Washington D.C.', 'USA', True
    dc.save()

    tok = MunicipalityM.new(doc_id="TOK")
    tok.city_name, tok.country, tok.capital = 'Tokyo', 'Japan', True
    tok.save()

    beijing = MunicipalityM.new(doc_id="BJ")
    beijing.city_name, beijing.country, beijing.capital = \
        'Beijing', 'China', True
    beijing.save()


@pytest.mark.usefixtures("setup_cities")
def test_subclass_same_collection(CTX):
    """
    Tests that subclasses of a class can be stored in the same collection.
    :return:
    """

    expected_dict = {
        'Washington D.C.': {
            'cityName': 'Washington D.C.',
            'country': 'USA',
            'capital': True,
            'obj_type': "MunicipalityM",
            'doc_id': 'DC',
            'doc_ref': 'CityM/DC'
        },
        'San Francisco': {
            'cityName': 'San Francisco',
            'cityState': 'CA',
            'country': 'USA',
            'capital': False,
            'regions': ['west_coast', 'norcal'],
            'obj_type': "StandardCityM",
            'doc_id': 'SF',
            'doc_ref': 'CityM/SF'
        },
        'Los Angeles': {
            'cityName': 'Los Angeles',
            'cityState': 'CA',
            'country': 'USA',
            'capital': False,
            'regions': ['west_coast', 'socal'],
            'obj_type': "StandardCityM",
            'doc_id': 'LA',
            'doc_ref': 'CityM/LA'
        }

    }

    res_dict = dict()

    for obj in CityM.where(CityM.country=="USA"):
        d = obj.to_dict()
        print(d)
        res_dict[d["cityName"]] = d

    assert res_dict['Washington D.C.'] == expected_dict['Washington D.C.']
    assert res_dict['San Francisco'] == expected_dict['San Francisco']
    assert res_dict['Los Angeles'] == expected_dict['Los Angeles']


@pytest.mark.usefixtures("setup_cities")
def test_subclass_query(CTX):
    assert {city.doc_id for city in CityM.all()} == \
           {'SF', 'LA', 'TOK', 'DC', 'BJ'}
    assert {city.doc_id for city in MunicipalityM.all()} == {'TOK', 'DC', 'BJ'}


@pytest.mark.usefixtures("setup_cities")
def test_where_with_kwargs(CTX):
    """
    Tests that subclasses of a class can be stored in the same collection.
    :return:
    """

    expected_dict = {
        'Washington D.C.': {
            'cityName': 'Washington D.C.',
            'country': 'USA',
            'capital': True,
            'obj_type': "MunicipalityM",
            'doc_id': 'DC',
            'doc_ref': 'CityM/DC'
        },
        'San Francisco': {
            'cityName': 'San Francisco',
            'cityState': 'CA',
            'country': 'USA',
            'capital': False,
            'regions': ['west_coast', 'norcal'],
            'obj_type': "StandardCityM",
            'doc_id': 'SF',
            'doc_ref': 'CityM/SF'
        },
        'Los Angeles': {
            'cityName': 'Los Angeles',
            'cityState': 'CA',
            'country': 'USA',
            'capital': False,
            'regions': ['west_coast', 'socal'],
            'obj_type': "StandardCityM",
            'doc_id': 'LA',
            'doc_ref': 'CityM/LA'
        }

    }

    res_dict = dict()

    for obj in CityM.where(country="USA"):
        d = obj.to_dict()
        res_dict[obj.city_name] = d

    assert res_dict['Washington D.C.'] == expected_dict['Washington D.C.']
    assert res_dict['San Francisco'] == expected_dict['San Francisco']
    assert res_dict['Los Angeles'] == expected_dict['Los Angeles']


@pytest.fixture
def setup_and_finalize(request, CTX):

    # Clear City collection before test
    _delete_all(collection_name="CityM", CTX=CTX)

    # Clear City collection after test
    def fin():
        _delete_all(collection_name="CityM", CTX=CTX)

    request.addfinalizer(fin)


@pytest.mark.usefixtures("setup_and_finalize")
def test_from_dict(CTX):
    sf = StandardCityM.from_dict(
        {
            'cityName': 'San Francisco',
            'cityState': 'CA',
            'country': 'USA',
            'capital': False,
            'regions': ['west_coast', 'norcal'],
        },
    )
    assert sf.city_state == "CA"
    sf.save()


def test_to_dict(CTX):
    sf = StandardCityM.new(doc_id="SF")
    sf.city_name, sf.city_state, sf.country, sf.capital, sf.regions = \
        'San Francisco', 'CA', 'USA', False, ['west_coast', 'norcal']
    assert sf.to_dict() == {
        'cityName': 'San Francisco',
        'cityState': 'CA',
        'country': 'USA',
        'capital': False,
        'regions': ['west_coast', 'norcal'],
        'doc_id': 'SF',
        'doc_ref': 'CityM/SF',
        'obj_type': 'StandardCityM'
    }


@pytest.mark.usefixtures("setup_and_finalize")
def test_from_dict_and_doc_id(CTX):
    sf = StandardCityM.from_dict(
        {
            'cityName': 'San Francisco',
            'cityState': 'CA',
            'country': 'USA',
            'capital': False,
            'regions': ['west_coast', 'norcal'],
        }, doc_id="SF"
    )

    sf.save()

    assert sf.doc_id == "SF"
    assert sf.doc_ref == "CityM/SF"

    # tear down steps
    sf.delete()




class BoilerError(Exception):
    """ The base for errors generated by flask-boiler.
    """
    pass


class ModelDeclarationError(BoilerError):
    pass


class DefaultNotAllowedError(BoilerError):
    """
    Thrown when a field's value is not provided and cannot
        be initialized with the default value.
    """
    pass


class PropertyEvalError(BoilerError):
    """ An error generated when getting a property from an instance
            of a model.
    """
    pass


class UnauthorizedError(BoilerError):
    pass



from collections import defaultdict


class ModelRegistry(type):
    """

    TODO: add exception handling when resolving classes
            that are destructed.

    Attributes:
    ===================
    _REGISTRY: dict
        key: name of the class
        value: class

    """

    _REGISTRY = {}
    _tree = defaultdict(set)
    _tree_r = defaultdict(set)

    def __new__(mcs, name, bases, attrs):
        new_cls = type.__new__(mcs, name, bases, attrs)
        if new_cls.__name__ in mcs._REGISTRY:
            raise ValueError(
                "Class with name {} is declared more than once. "
                .format(new_cls.__name__)
            )
        mcs._REGISTRY[new_cls.__name__] = new_cls

        for base in bases:
            if issubclass(type(base), ModelRegistry):
                mcs._tree[base.__name__].add(new_cls.__name__)
                mcs._tree_r[new_cls.__name__].add(base.__name__)

        return new_cls

    @classmethod
    def get_registry(mcs):
        return dict(mcs._REGISTRY)

    @classmethod
    def _get_children_str(mcs, cls_name):
        return mcs._tree[cls_name].copy()

    @classmethod
    def _get_parents_str(mcs, cls_name):
        return mcs._tree_r[cls_name].copy()

    @classmethod
    def get_cls_from_name(mcs, obj_type_str):
        """ Returns cls from obj_type (classname string)
        obj_type must be a subclass of cls in current class/object
        """
        if obj_type_str in mcs._REGISTRY:
            return mcs._REGISTRY[obj_type_str]
        else:
            return None


import collections
import abc

from onto.common import _NA

import abc
from enum import Enum, auto


class ConditionArg(Enum):
    """
    Do not declare attributes here
    """
    pass


class Reference(collections.UserString):
    """
    example: '/myusername/'
    """

    def __init__(self, _s='', *, _is_empty=True):
        """
        TODO: make compatible with sequence input
        :param _s:
        :param _is_empty:
        """
        if not isinstance(_is_empty, bool):
            import logging
            logging.error(
                f'Expected bool, but received {_is_empty} '
                f'as the value for _is_empty. '
                'Initialize with Reference.from_str where applicable. '
            )
        self._is_empty = _is_empty
        super().__init__(_s)

    def child(self, s):
        if self._is_empty:
            return self.__class__(_s=s, _is_empty=False)
        else:
            return self.__class__(_s=f'{self}/{s}', _is_empty=False)

    @classmethod
    def from_str(cls, s: str):
        if s == '':
            return cls(_is_empty=True, _s=s)
        else:
            return cls(_is_empty=False, _s=s)

    @property
    def first(self):
        return self.data.split('/')[0]  # example: '/myusername/'

    @property
    def last(self):
        return self.data.split('/')[-1]

    @property
    def id(self):
        return self.last

    @property
    def params(self):
        return self.data.split('/')

    @property
    def to_str(self):
        return str(self)

    @property
    def path(self) -> tuple:
        return tuple(self.params)

    @property
    def collection(self):
        raise NotImplementedError

    def __truediv__(self, other):
        """ Overload / operator.
            Example: SomeReference / 'new_path_segment'

        :param other:
        :return:
        """
        return self.child(other)

    def __rtruediv__(self, other):
        raise TypeError('str / Reference is not supported')


reference = Reference()


class Snapshot(collections.UserDict):

    next = property()
    prev = property()

    def __init__(self, *args, __onto_meta__=None, **kwargs):
        """
        TODO: note that in extreme conditions, a document with randomly
        TODO:   generated string key may take on the value of
        TODO:   '__onto_meta__' and thus cause error

        :param args:
        :param __onto_meta__:
        :param kwargs:
        """
        super().__init__(*args, **kwargs)
        if __onto_meta__ is not None:
            for key, val in __onto_meta__.items():
                setattr(self, key, val)

    @next.getter
    def next(self):
        return self._onto_next

    @next.setter
    def next(self, _onto_next):
        self._onto_next = _onto_next

    @prev.getter
    def prev(self):
        return self._onto_prev

    @prev.setter
    def prev(self, _onto_prev):
        self._onto_prev = _onto_prev

    def to_dict(self):
        return self.data.copy()


class Database:

    class Comparators(ConditionArg):
        @property
        def condition(self):
            return self.value

    @classmethod
    @abc.abstractmethod
    def set(cls, ref: Reference, snapshot: Snapshot, transaction=_NA):
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def get(cls, ref: Reference, transaction=_NA):
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def get_many(cls, refs: [Reference], transaction=_NA):
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def update(cls, ref: Reference, snapshot: Snapshot, transaction=_NA):
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def create(cls, ref: Reference, snapshot: Snapshot, transaction=_NA):
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def delete(cls, ref: Reference, transaction=_NA):
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def query(cls, q):
        raise NotImplementedError

    ref = reference

    # TODO: NOTE: Should be classmethod
    listener = None


class Listener:

    _registry = dict()
    from ..coordinator import Coordinator
    _coordinator = Coordinator()

    @classmethod
    def for_query(cls):
        pass

    @classmethod
    def for_refs(cls):
        pass


def is_reference(val):
    return issubclass(val.__class__, Reference)


def is_snapshot(val):
    return issubclass(val.__class__, Snapshot)



import time
from functools import partial
from pathlib import Path
from subprocess import Popen

from pytest_kafka import (
    make_zookeeper_process, make_kafka_server, make_kafka_consumer,
    terminate,
)

from onto.database.kafka import KafkaDatabase


def get_bins():
    from pathlib import Path
    ROOT = Path.cwd()
    KAFKA_SCRIPTS = ROOT / 'kafka_2.12-2.7.0/bin/'
    KAFKA_BIN = str(KAFKA_SCRIPTS / 'kafka-server-start.sh')
    ZOOKEEPER_BIN = str(KAFKA_SCRIPTS / 'zookeeper-server-start.sh')
    # You can pass a custom teardown function (or parametrise ours). Just don't call it `teardown`
    # or Pytest will interpret it as a module-scoped teardown function.
    return KAFKA_BIN, ZOOKEEPER_BIN


KAFKA_BIN, ZOOKEEPER_BIN = get_bins()

# You can pass a custom teardown function (or parametrise ours). Just don't call it `teardown`
# or Pytest will interpret it as a module-scoped teardown function.
teardown_fn = partial(terminate, signal_fn=Popen.kill)
zookeeper_proc = make_zookeeper_process(ZOOKEEPER_BIN, teardown_fn=teardown_fn)
kafka_server = make_kafka_server(KAFKA_BIN, 'zookeeper_proc', kafka_port=9092,
                                 teardown_fn=teardown_fn)
kafka_consumer = make_kafka_consumer(
    'kafka_server', seek_to_beginning=True, kafka_topics=['topic'])

def test_fixture(kafka_server):
    from kafka import KafkaProducer

    # assert isinstance(kafka_server, KafkaProducer)
    time.sleep(5)

    def value_serializer(x):
        import json
        return json.dumps(x).encode('utf-8')

    kafka_server = KafkaProducer(bootstrap_servers='localhost:9092', value_serializer=value_serializer)
    kafka_server.send('SelectedStudent', key='mykey'.encode('utf-8'), value={'hello': 'world'})
    time.sleep(3)


def test_start(kafka_server):
    from onto.source.kafka import KafkaSource
    src = KafkaSource(topic_name='my-topic')
    import asyncio
    loop = asyncio.get_event_loop()
    src.start(loop=loop)

    # Send something on Kafka
    from kafka import KafkaProducer
    kafka_server = KafkaProducer(bootstrap_servers='localhost:9092')
    kafka_server.send('my-topic', value='hello world!'.encode('utf-8'))

    # # Run trivial uvicorn
    # import asyncio
    # loop = asyncio.get_event_loop()
    # from starlette.applications import Starlette
    # app = Starlette()
    # from uvicorn import Config
    # config = Config(app=app, loop=loop, port=8080, debug=True)
    # from uvicorn import Server
    # server = Server(config)
    # loop.run_until_complete(server.serve())


def test__register():
    from unittest.mock import patch
    import onto.source.kafka

    async def _mock_subscribe():
        """
        https://docs.python.org/3/library/asyncio-task.html
        :return:
        """
        import asyncio
        import datetime
        loop = asyncio.get_running_loop()
        end_time = loop.time() + 5.0
        while True:
            print(datetime.datetime.now())
            if (loop.time() + 1.0) >= end_time:
                break
            await asyncio.sleep(1)
            yield datetime.datetime.now()

    with patch.object(onto.source.kafka, '_kafka_subscribe', _mock_subscribe):
        from onto.source.kafka import KafkaSource
        src = KafkaSource(topic_name='my-topic')
        import asyncio
        loop = asyncio.get_event_loop()
        src.start(loop=loop)

        # import asyncio
        # loop = asyncio.get_event_loop()
        # from starlette.applications import Starlette
        # app = Starlette(debug=True)
        # from uvicorn import Config
        # config = Config(app=app, loop=loop, port=8080, debug=True)
        # from uvicorn import Server
        # server = Server(config)
        # loop.run_until_complete(server.serve())


def test_save_instance(kafka_server):
    # from examples.meeting_room.domain_models import Meeting

    from pony.orm import sql_debug
    from onto.collection_mixin import db
    from pony.orm.dbproviders.sqlite import SQLiteProvider

    import pony.options
    pony.options.CUT_TRACEBACK = False

    from examples.meeting_room.domain_models import Meeting
    sql_debug(True)  # Output all SQL queries to stdout

    db.bind(SQLiteProvider, filename=':memory:', create_db=True)
    db.generate_mapping(check_tables=False, create_tables=True)
    from pony.orm.core import DBSessionContextManager
    db_session = DBSessionContextManager()
    time.sleep(5)

    @db_session
    def f():
        m = Meeting.new(doc_id='mm')
        r = Meeting._datastore().kafka_producer().send(topic='meetings', key='hello'.encode('utf-8'), value='world'.encode('utf-8'))
        m.save()
        time.sleep(2)

    f()



from .domain_model import DomainModelSource as domain_model, \
    DomainModelTransactionalSource as domain_model_transactional
from .firestore import FirestoreSource as firestore
from .view_model import ViewModelSource as view_model
from .rest import RestViewModelSource as rest




if __name__ == "__main__":

    from onto.attrs.unit import _ModelRegistry

    classnames = set(_ModelRegistry._REGISTRY.keys())

    import ast

    with open('onto/attrs/unit.py') as unit:
        tree = ast.parse(unit.read())


        def yield_elems():
            for elem in tree.body:
                if isinstance(elem, ast.ClassDef):
                    name = elem.name
                    if name in classnames:
                        yield elem

            else:
                yield from ()


        def get_init(elem: ast.ClassDef):
            init = None

            class CollectInit(ast.NodeVisitor):

                def visit_FunctionDef(self, node: ast.FunctionDef):
                    if node.name == '__init__':
                        nonlocal init
                        init = node

            CollectInit().visit(elem)

            return init


        with open('onto/attrs/attribute_new.py') as attribute_new:
            attribute_new_tree = ast.parse(attribute_new.read())
            attribute_new_class: ast.ClassDef = next(elem for elem in attribute_new_tree.body if
                                                     isinstance(elem, ast.ClassDef) and elem.name == 'AttributeBase')
            # attribute_new_clas

        for elem in yield_elems():
            init = get_init(elem)
            if init is None:
                continue
            from inflection import underscore

            name = underscore(elem.name)
            new_f = ast.FunctionDef(name=name, args=init.args, body=[ast.Pass()], decorator_list=[],
                                    returns=[ast.Constant(value='AttributeBase')])

            # f = ast.unparse(new_f)
            # print(f)

            attribute_new_class.body.append(
                new_f
            )

        with open("onto/attrs/attribute_new.pyi", 'w+') as io:
            attribute_new_tree = ast.fix_missing_locations(attribute_new_tree)
            ss = ast.unparse(attribute_new_tree)
            io.write(ss)



import time

from examples.meeting_room.view_models import MeetingSession
from examples.meeting_room.view_models.meeting_session import \
    MeetingSessionMutation
from ..views import meeting_session_ops
from onto import testing_utils
from onto.view import rest_api
from .. import view_models
# Import the fixtures used by fixtures
from tests.fixtures import CTX, setup_app
from .fixtures import users, tickets, location, meeting


def test_view_model(users, tickets, location, meeting):
    meeting_session = view_models.MeetingSession \
        .get(doc_id=meeting.doc_id, once=True)

    assert meeting_session.to_dict() == \
           {'inSession': True,
            'longitude': -117.242929,
            'latitude': 32.880361,
            'address': '9500 Gilman Drive, La Jolla, CA',
            'attending': [
                {
                    'name': 'Joshua Pendergrast',
                    'organization': 'SDSU',
                    'hearing_aid_requested': True},
                {
                    'name': 'Thomasina Manes',
                    'organization': 'UCSD',
                    'hearing_aid_requested': False},
                {
                    'name': 'Tijuana Furlong',
                    'organization': 'UCSD',
                    'hearing_aid_requested': True},
            ],
            'numHearingAidRequested': 2}


def test_view(users, tickets, location, meeting, setup_app):
    app = setup_app

    mediator = rest_api.ViewMediator(
        view_model_cls=MeetingSession,
        app=app,
        mutation_cls=MeetingSessionMutation
    )
    mediator.add_list_get(
        rule="/meeting_sessions",
        list_get_view=meeting_session_ops.ListGet
    )

    testing_utils._wait(factor=2)

    test_client = app.test_client()

    res = test_client.get(
        path='meeting_sessions?status=in-session')

    assert res.json == {'results': {
        "meeting_1": {'inSession': True,
         'longitude': -117.242929,
         'latitude': 32.880361,
         'address': '9500 Gilman Drive, La Jolla, CA',
         'attending': [
             {
                 'name': 'Joshua Pendergrast',
                 'organization': 'SDSU',
                 'hearing_aid_requested': True},
             {
                 'name': 'Thomasina Manes',
                 'organization': 'UCSD',
                 'hearing_aid_requested': False},
             {
                 'name': 'Tijuana Furlong',
                 'organization': 'UCSD',
                 'hearing_aid_requested': True},
         ],
         'numHearingAidRequested': 2}
    }
    }

    mediator.add_instance_get(rule="/meeting_sessions/<string:doc_id>")
    res = test_client.get(
        path='meeting_sessions/meeting_1')

    assert res.json == {'inSession': True,
         'longitude': -117.242929,
         'latitude': 32.880361,
         'address': '9500 Gilman Drive, La Jolla, CA',
         'attending': [
             {
                 'name': 'Joshua Pendergrast',
                 'organization': 'SDSU',
                 'hearing_aid_requested': True},
             {
                 'name': 'Thomasina Manes',
                 'organization': 'UCSD',
                 'hearing_aid_requested': False},
             {
                 'name': 'Tijuana Furlong',
                 'organization': 'UCSD',
                 'hearing_aid_requested': True},
         ],
         'numHearingAidRequested': 2}

    mediator.add_instance_patch(rule="/meeting_sessions/<string:doc_id>")
    test_client.patch(
        path='meeting_sessions/meeting_1', json={
            "inSession": False
        })

    testing_utils._wait(factor=.7)

    res = test_client.get(
        path='meeting_sessions/meeting_1')

    assert res.json == {'inSession': False,
                        'longitude': -117.242929,
                        'latitude': 32.880361,
                        'address': '9500 Gilman Drive, La Jolla, CA',
                        'attending': [
                            {
                                'name': 'Joshua Pendergrast',
                                'organization': 'SDSU',
                                'hearing_aid_requested': True},
                            {
                                'name': 'Thomasina Manes',
                                'organization': 'UCSD',
                                'hearing_aid_requested': False},
                            {
                                'name': 'Tijuana Furlong',
                                'organization': 'UCSD',
                                'hearing_aid_requested': True},
                        ],
                        'numHearingAidRequested': 2}


def test_user_view(users, meeting):
    user_view = view_models.UserView.get_from_user_id(user_id="thomasina", )

    # time.sleep(2)  # TODO: delete after implementing sync

    assert user_view.to_dict() == {'meetings': [
        {'status': 'in-session', 'users': [
            {'lastName': 'Furlong', 'firstName': 'Tijuana',
             'hearingAidRequested': True, 'organization': 'UCSD'},
            {'lastName': 'Manes', 'firstName': 'Thomasina',
             'hearingAidRequested': False, 'organization': 'UCSD'},
            {'lastName': 'Pendergrast', 'firstName': 'Joshua',
             'hearingAidRequested': True, 'organization': 'SDSU'}],
         'location': {'latitude': 32.880361,
                      'address': '9500 Gilman Drive, La Jolla, CA',
                      'longitude': -117.242929},
         'tickets': {'joshua': {'attendance': True,
                                      'role': 'Participant',
                                      'user': {'firstName': 'Joshua',
                                               'hearingAidRequested': True,
                                               'lastName': 'Pendergrast',
                                               'organization': 'SDSU'}},
                           'thomasina': {'attendance': True,
                                         'role': 'Organizer',
                                         'user': {'firstName': 'Thomasina',
                                                  'hearingAidRequested': False,
                                                  'lastName': 'Manes',
                                                  'organization': 'UCSD'}},
                           'tijuana': {'attendance': True,
                                       'role': 'Participant',
                                       'user': {'firstName': 'Tijuana',
                                                'hearingAidRequested': True,
                                                'lastName': 'Furlong',
                                                'organization': 'UCSD'}}}}],
        'lastName': 'Manes',
        'firstName': 'Thomasina',
        'hearingAidRequested': False,
        'organization': 'UCSD'}


def test_view_model_update(users, tickets, location, meeting):
    meeting_session = view_models.MeetingSession \
        .get(doc_id=meeting.doc_id, once=False)

    # time.sleep(2)  # TODO: delete after implementing sync

    tickets[0].attendance = False
    tickets[0].save()

    testing_utils._wait()

    assert meeting_session.to_dict() == {'inSession': True,
                                                      'longitude': -117.242929,
                                                      'latitude': 32.880361,
                                                      'address': '9500 Gilman Drive, La Jolla, CA',
                                                      'attending': [
                                                          {
                                                              'name': 'Joshua Pendergrast',
                                                              'organization': 'SDSU',
                                                              'hearing_aid_requested': True},
                                                          {
                                                              'name': 'Thomasina Manes',
                                                              'organization': 'UCSD',
                                                              'hearing_aid_requested': False}
                                                      ],
                                                      'numHearingAidRequested': 1}



import pytest

from onto import collection_mixin, domain_model
from abc import ABC
from .fixtures import CTX


def test_init():

    class ExampleMixin(collection_mixin.CollectionMixin):
        _collection_name = "Examples"

    assert ExampleMixin._collection_name == "Examples"


@pytest.fixture
def ExampleMixin(CTX):

    class ExampleMixin(collection_mixin.CollectionMixin):
        _collection_name = "Examples"

        @classmethod
        def _datastore(cls):
            return CTX.dbs.firestore

    return ExampleMixin


def test_get_collection(CTX, ExampleMixin):
    assert ExampleMixin._get_collection().id == "Examples"


def test_get_collection_name(CTX, ExampleMixin):
    assert ExampleMixin._get_collection_name() == "Examples"


def test_mixin_instance_methods(CTX, ExampleMixin):
    example = ExampleMixin()
    assert example.collection_name == "Examples"
    assert example.collection.id == "Examples"


def test_mix_metaclass(CTX):

    class ExampleMetaclassObject(ABC):

        example_static_attr = "example_static_attr"

        def example_method(self):
            return "example_method"

        @classmethod
        def example_cls_method(cls):
            return "example_cls_method"

    class ExampleMixedClass(ExampleMetaclassObject,
                            collection_mixin.CollectionMixin,
                            ):
        _collection_name = "Examples"

    assert ExampleMixedClass._collection_name == "Examples"

    assert ExampleMixedClass.example_static_attr == "example_static_attr"
    assert ExampleMixedClass.example_cls_method() == "example_cls_method"

    example = ExampleMixedClass()
    assert example.example_method() == "example_method"




from onto import schema, fields, domain_model, factory


class PostSchema(schema.Schema):
    owner_id = fields.Str()


class PostBase(domain_model.DomainModel):

    class Meta:
        collection_name = "insta_posts"


class Post(PostBase):

    class Meta:
        schema_cls = PostSchema



from onto import domain_model
from onto.attrs import attrs
# from onto.collection_mixin import with_pony


class UserBase(domain_model.DomainModel
    # , metaclass=with_pony
               ):
    # doc_ref = attrs.doc_ref(type_cls=str)
    pass


class User(UserBase):

    class Meta:
        collection_name = "users"

    first_name = attrs.str
    last_name = attrs.str
    organization = attrs.str
    hearing_aid_requested = attrs.str
    display_name = attrs.str.optional

    meetings = attrs.list(
        value=lambda a: a.relation('Meeting')
    ).optional.default_value(list)
    tickets = attrs.list(
        value=lambda a: a.relation('Tickets')
    ).optional.default_value(list)


    @classmethod
    def _datastore(cls):
        from onto.database.mock import MockDatabase
        return MockDatabase

    @display_name.getter
    def display_name(self):
        return "{} {}".format(self.first_name, self.last_name)



from .base import Sink

import flask


class RestServerSink(Sink):

    def emit_and_close(self, response: flask.Response):
        flask.abort(response)


class RestClientSink(Sink):

    def emit(self, url, data):
        import requests
        requests.post(url=url, data=data)

rest_server = RestServerSink

rest_client = RestClientSink



import unittest

from onto.database.firestore import FirestoreDatabase
from onto.query import cmp


class MyTestCase(unittest.TestCase):

    def test_eq(self):
        a = cmp.CMP().age
        b = a == 1
        assert isinstance(b, cmp.NodeCondition)
        assert b.constraints == [(FirestoreDatabase.Comparators.eq, 1)]

    def test_range(self):
        a = cmp.CMP().age
        b = 0 <= a <= 1
        assert isinstance(b, cmp.NodeCondition)
        assert b.constraints == [(FirestoreDatabase.Comparators.ge, 0), (FirestoreDatabase.Comparators.le, 1)]

    def test_excl_range(self):
        a = cmp.CMP().day
        b = 0 < a < 8
        assert isinstance(b, cmp.NodeCondition)
        assert b.constraints == [(FirestoreDatabase.Comparators.gt, 0), (FirestoreDatabase.Comparators.lt, 8)]

    def test_neq_raise(self):
        a = cmp.CMP().day

        def p():
            _ = a != 5

        self.assertRaises(ValueError, p)

    def test_in(self):
        a = cmp.CMP().friends
        assert isinstance(a, cmp.RootCondition)
        b = a.has("user_k")
        assert isinstance(b, cmp.NodeCondition)
        assert b.constraints == [(FirestoreDatabase.Comparators.contains, "user_k")]


if __name__ == '__main__':
    unittest.main()



from unittest import mock, skip

import pytest
from testfixtures import compare

import onto.utils
from onto import schema, fields
from onto.errors import PropertyEvalError
from onto.mapper.schema import Schema, BoilerProperty
from onto.models.base import Serializable


def test__get_instance_vars():
    class TestObject:

        def __init__(self):
            self.int_a = 0
            self.int_b = 0

    obj_cls = TestObject

    res = schema._get_instance_variables(obj_cls)

    assert res == ["int_a", "int_b"]


@pytest.mark.skip
def test__get_instance_vars_nested_init():
    class TestObject:

        def _nested_init(self):
            self.int_c = 0

        def __init__(self):
            self.int_a = 0
            self.int_b = 0
            self._nested_init()

    obj_cls = TestObject

    res = schema._get_instance_variables(obj_cls)
    # print(res)
    assert res == ["int_a", "int_b", "int_c"]


@pytest.mark.skip
def test__get_field_vars():
    """
    Tests that _get_field_vars returns a dictionary of expected Field objects.
    """

    # class TestObject:
    #
    #     def __init__(self):
    #         self.int_a = 0
    #         self.int_b = 0

    # obj_cls = TestObject

    field_vars = schema._get_field_vars(["int_a", "int_b"],
                                        fd=["int_a", "int_b"])

    def field_comparator(x, y, _):
        """
        Test fixture for comparing two fields.Field objects
        """
        if x.load_from != y.load_from:
            return "x != y since x.load_from: {} != y.load_from: {}".format(
                x.load_from, y.load_from
            )

    compare(

        field_vars, {
            "int_a": fields.Raw(load_from="intA"),
            "int_b": fields.Raw(load_from="intB"),
        }, comparers={fields.Field: field_comparator}

    )


def test_set_attr():
    """
    Tests that generate_schema returns a schema that has the ability to
        set instance variables based on keys of different format in the
        dictionary provided in schema.load(d)
    """

    class TestObject(object):

        def __init__(self):
            super().__init__()
            self.int_a = 0
            self.int_b = 0

        @classmethod
        def get_fields(cls):
            return ["int_a", "int_b"]

        def __new__(cls, *args, **kwargs):
            instance = super().__new__(cls, *args, **kwargs)
            return instance

    s = schema.generate_schema(TestObject)

    # print(s)

    obj = s.load({
        "intA": 1,
        "intB": 2
    })

    assert obj.int_a == 1
    assert obj.int_b == 2

    # TODO: test that no extraneous attributes are set


def test_conversion():
    """
    Tests that "camelCase" key used in Firestore converts to "snake_case"
        target instance variable names.
    """

    res = onto.utils.firestore_key_to_attr_name("intA")
    assert res == "int_a"

    r_res = onto.utils.attr_name_to_firestore_key("int_a")
    assert r_res == "intA"


@pytest.fixture
def location_schema_fields():

    coordinates_field = fields.Raw()
    coordinates_field.deserialize = mock.MagicMock()
    address_field = fields.Raw()
    address_field.deserialize = mock.MagicMock()

    class LocationSchema(schema.Schema):
        coordinates = coordinates_field
        address = address_field

    return LocationSchema(), coordinates_field, address_field


def test_raw():
    coordinates_field = fields.Raw()
    address_field = fields.Raw()

    class LocationSchema(schema.Schema):
        coordinates = coordinates_field
        address = address_field

    location_schema = LocationSchema()

    serialized = {
        "coordinates": {
            "latitude": 36.98765,
            "longitude": 100.12345,
        },

        "address": "9500 Gilman Dr."
    }

    mres = location_schema.load(serialized)

    deserialized_dict = mres

    assert deserialized_dict == serialized


def test_schema_new():
    from onto.primary_object import PrimaryObjectSchema
    class CitySchema(PrimaryObjectSchema):
        city_name = fields.Raw(data_key="name")

        country = fields.Raw(data_key="country")
        capital = fields.Raw(data_key="capital")

    class MunicipalitySchema(CitySchema):
        pass

    class StandardCitySchema(CitySchema):
        city_state = fields.Raw(data_key="state")
        regions = fields.Raw(many=True, data_key="regions")

    assert set(CitySchema().fields.keys()) == {
        "city_name", "country", "capital",
        "doc_id", "doc_ref", "obj_type", "_remainder"
    }

    assert set(MunicipalitySchema().fields.keys()) == {
        "city_name", "country", "capital",
        "doc_id", "doc_ref", "obj_type", "_remainder"
    }

    assert set(StandardCitySchema().fields.keys()) == {
        "city_state", "regions",
        "city_name", "country", "capital",
        "doc_id", "doc_ref", "obj_type", "_remainder"
    }


def test_schema_dump():

    class CitySchema(Schema):
        city_name = fields.Raw()

        country = fields.Raw()
        capital = fields.Raw()

    class MunicipalitySchema(CitySchema):
        pass

    class StandardCitySchema(CitySchema):
        city_state = fields.Raw()
        regions = fields.Raw(many=True)

    assert MunicipalitySchema().dump({
        'city_name': 'Washington D.C.',
        'country': 'USA',
        'capital': True,
        # 'obj_type': "Municipality",
        # 'doc_id': 'DC',
        # 'doc_ref': 'City/DC'
    }) == {
        'cityName': 'Washington D.C.',
        'country': 'USA',
        'capital': True,
        'obj_type': "dict",
        # 'doc_id': 'DC',
        # 'doc_ref': 'City/DC'
    }


def test_schema_load():

    class CitySchema(Schema):
        city_name = fields.Raw()

        country = fields.Raw()
        capital = fields.Raw()

    class MunicipalitySchema(CitySchema):
        pass

    class StandardCitySchema(CitySchema):
        city_state = fields.Raw()
        regions = fields.Raw(many=True)

    assert MunicipalitySchema().load({
        'cityName': 'Washington D.C.',
        'country': 'USA',
        'capital': True,
    }) == {
        'city_name': 'Washington D.C.',
        'country': 'USA',
        'capital': True,
    }


def test_property_get_error_handling():

    class CitySchemaExtended(Schema):
        city_name = fields.Raw()

        country = fields.Raw()
        capital = fields.Raw()

        extra_property = fields.Raw(dump_only=True)

    class CityExtra(Serializable):

        _schema_cls = CitySchemaExtended

        @BoilerProperty
        def extra_property(self):
            return self.non_existent_property

    city = CityExtra.new(city_name="San Diego", country="USA", capital=False)

    """
    Tests that AttributeError is not swallowed when 
        the property function is located but failed 
        during evaluation 
    """
    with pytest.raises(PropertyEvalError) as excinfo:
        res = CitySchemaExtended().dump(city)

    assert isinstance(excinfo.value.__cause__, AttributeError)

    """
    Tests that AttributeError is raised when accessing 
        a property that does not exist 
    TODO: move to test_serializable 
    """
    with pytest.raises(AttributeError):
        city.non_existent_property
